
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu


  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/





#include "PLA.h"
#include "MISC_DEFS.h"

extern  MPI_Op PLA_SUM;

static PLA_Obj alpha_in = NULL, a_in = NULL, x_in = NULL,
					beta_in = NULL, y_in = NULL;

int PLA_Symv_enter ( int uplo, PLA_Obj alpha, PLA_Obj a, PLA_Obj x,
						  PLA_Obj beta,  PLA_Obj y )
{
	int check_parameter, instrument, check_seq, myrow, mycol;
	char current_routine[20] = "PLA_Symv";
	int objtype;
	MPI_Datatype datatype;
	PLA_Template template;
	int  error = 0, original_error = 0;

	PLA_Stack_push_current( current_routine );

	PLA_Get_check_parameters ( &check_parameter );

	if ( PLA_CHECK_PARAMETER == check_parameter )
	{
	/* check uplo */
		if ( PLA_UP_TRIAN != uplo && PLA_LOW_TRIAN != uplo )
			PLA_Log_abort("PLA_Symv_enter ", " PLA_Symv : uplo invalid", PLA_ERROR_CODE, original_error);

	/* check objects for validity */
		PLA_Obj_object_valid( alpha, "alpha", error, original_error);
		PLA_Obj_object_valid( a, "a", error, original_error);
		PLA_Obj_object_valid( x, "x", error, original_error);
		PLA_Obj_object_valid( beta, "beta", error, original_error);
		PLA_Obj_object_valid( y, "y", error, original_error);

	 /* check object types */
		PLA_Obj_objtype_equal( alpha, PLA_MSCALAR, "alpha", error, original_error );
		PLA_Obj_objtype_equal( a,     PLA_MATRIX,  "a", error, original_error );
		PLA_Obj_objtype_equal( x,     PLA_ANY_VECTOR, "x", error, original_error );
		PLA_Obj_objtype_equal( beta,  PLA_MSCALAR, "beta", error, original_error );
		PLA_Obj_objtype_equal( y,     PLA_ANY_VECTOR,  "y", error, original_error );

	 /* check datatypes */
		PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
		PLA_Obj_datatype( alpha, &datatype );
		PLA_Obj_datatype_equal( a,     datatype, "a", error, original_error );
		PLA_Obj_datatype_equal( x,     datatype, "x", error, original_error );
		PLA_Obj_datatype_equal( beta,  datatype, "beta", error, original_error );
		PLA_Obj_datatype_equal( y,     datatype, "y", error, original_error );

	 /* check if templates are all equal (temporary restriction) */
		PLA_Obj_template( alpha, &template );
		PLA_Obj_template_equal( a,     template, "a", error, original_error );
		PLA_Obj_template_equal( x,     template, "x", error, original_error );
		PLA_Obj_template_equal( beta,  template, "beta", error, original_error );
		PLA_Obj_template_equal( y,     template, "y", error, original_error );

	 /* check dimensions */
		PLA_Obj_global_length_equal ( alpha, 1, "alpha", error, original_error );
		PLA_Obj_global_width_equal  ( alpha, 1, "alpha", error, original_error );
		PLA_Obj_global_length_equal ( beta, 1, "beta", error, original_error );
		PLA_Obj_global_width_equal  ( beta, 1, "beta", error, original_error );

		PLA_Obj_global_width_length_equal ( a, x, "a and x", error, original_error );
		PLA_Obj_global_length_width_equal ( a, y, "a and y", error, original_error );
		PLA_Obj_global_length_width_equal ( a, a, "a is square", error, original_error);
	}
	PLA_Get_check_sequential ( &check_seq );
	if ( PLA_CHECK_SEQUENTIAL == check_seq )
	{
		 PLA_Mscalar_create_conf_to( alpha, 0, 0, &alpha_in );
		 PLA_Mscalar_create_conf_to( a,     0, 0, &a_in );
		 PLA_Mscalar_create_conf_to( x,     0, 0, &x_in );
		 PLA_Mscalar_create_conf_to( beta,  0, 0, &beta_in );
		 PLA_Mscalar_create_conf_to( y,     0, 0, &y_in );

		 PLA_Copy_temp( alpha, alpha_in );
		 PLA_Copy_temp( a,     a_in );
		 PLA_Copy_temp( x,     x_in );
		 PLA_Copy_temp( beta,  beta_in );
		 PLA_Copy_temp( y,     y_in );

		 PLA_Temp_comm_col_rank( template, &myrow );
		 PLA_Temp_comm_row_rank( template, &mycol );

		 if ( 0 == myrow && 0 == mycol )
			 PLA_Local_symv( uplo, alpha_in, a_in, x_in, beta_in, y_in );
	}

	PLA_Get_instrument ( &instrument );
	if ( PLA_INSTRUMENT == instrument )
	{
	  /* insert instrumentation calls */
	}
	return(original_error);
}


int PLA_Symv_exit ( int uplo,
						  PLA_Obj alpha, PLA_Obj a, PLA_Obj x,
						  PLA_Obj beta,  PLA_Obj y )
{
  char current_routine[20];
  int  instrument, check_seq, myrow, mycol;
  PLA_Obj y_out = NULL;
  PLA_Template template;

  PLA_Get_check_sequential ( &check_seq );
/* MAY (JAG) want to test the objects for validity here as well      */
/* in case we have a wayward pointer or some other memory corruption */
  if ( PLA_CHECK_SEQUENTIAL == check_seq )
  {
	  PLA_Mscalar_create_conf_to( y,     0, 0, &y_out );
	  PLA_Copy_temp(y, y_out);
	  PLA_Obj_template(y, &template);
	  PLA_Temp_comm_col_rank( template, &myrow );
	  PLA_Temp_comm_row_rank( template, &mycol );
	  if ( 0 == myrow && 0 == mycol )
	  {
/* Commented out because I know we have the code to do this somewhere */
/* and I would like to avoid re-writing it.  It's not a big deal to   */
/* rewrite if we restrict what can be passed to the function though.  */
/*        diff = PLA_max_abs_diff( c_in, c_out ); */
		}
	}

	PLA_Get_instrument ( &instrument );
	if ( PLA_INSTRUMENT == instrument )
	{
	  /* insert instrumentation calls */
	}
	PLA_Stack_pop_current( current_routine );
	/* Free all of the "in" objects */
	PLA_Obj_free( & alpha_in );
	PLA_Obj_free( & a_in     );
	PLA_Obj_free( & x_in     );
	PLA_Obj_free( & beta_in  );
	PLA_Obj_free( & y_in     );
	PLA_Obj_free( & y_out    );
	return(PLA_SUCCESS); /* Actually, return fail if exceed tolerance in test */
}

int  PLA_Symv_upper( PLA_Obj alpha, PLA_Obj a,
		PLA_Obj xtilde_row, PLA_Obj xtilde_col,
	        PLA_Obj one, PLA_Obj ytilde_row, PLA_Obj ytilde_col)
{
	PLA_Obj a_11 = NULL, a_12 = NULL, a_22 = NULL;
	PLA_Obj x_row_left = NULL,  x_row_right  = NULL;
	PLA_Obj x_col_top = NULL,  x_col_bottom  = NULL;
	PLA_Obj y_col_top  = NULL,  y_col_bottom = NULL;
	PLA_Obj y_row_left = NULL,   y_row_right = NULL;


	MPI_Datatype datatype;

	int     right_size, right_owner, bottom_size, bottom_owner, square_size;
	int     my_proc_row, my_proc_col;
	int     nb, rows_in_ret_buf, cols_in_ret_buf;
	PLA_Template template;

	PLA_Obj_template(a, &template);
	PLA_Temp_comm_col_rank( template, &my_proc_row );
	PLA_Temp_comm_row_rank( template, &my_proc_col );
	PLA_Temp_nb( template, &nb );
	PLA_Obj_datatype(a, &datatype);
	PLA_Obj_view_all(a, &a_11);
	PLA_Obj_view_all(xtilde_row, &x_row_left);
	PLA_Obj_view_all(xtilde_col, &x_col_top);
	PLA_Obj_view_all(ytilde_row, &y_row_left);
	PLA_Obj_view_all(ytilde_col, &y_col_top);

	while(TRUE)
	{
		PLA_Obj_split_size( a_11, PLA_SIDE_RIGHT, &right_size, &right_owner);
		PLA_Obj_split_size( a_11, PLA_SIDE_BOTTOM,  &bottom_size,  &bottom_owner );
		if( 0 == (square_size = min(right_size, bottom_size))) break;
		PLA_Obj_split_4(a_11, -square_size, -square_size, &a_11, &a_12,
							       PLA_DUMMY,&a_22);
		PLA_Obj_horz_split_2(y_col_top, -square_size, &y_col_top,
					  &y_col_bottom );
		PLA_Obj_horz_split_2(x_col_top, -square_size, &x_col_top,
				     &x_col_bottom);
		PLA_Obj_vert_split_2(x_row_left,  -square_size, &x_row_left, &x_row_right  );
		PLA_Obj_vert_split_2(y_row_left,  -square_size, &y_row_left, &y_row_right  );

/* resize the traspose buffer space */
		if (my_proc_col == right_owner)
		{
                        PLA_Obj_implicit_local_transpose(x_row_right);
                        PLA_Obj_implicit_local_transpose(y_row_right);
			if( my_proc_row == bottom_owner)
			   PLA_Local_symv( PLA_UPPER_TRIANGULAR, alpha, 
                              a_22, x_row_right, one, y_col_bottom);
			PLA_Local_gemv( PLA_TRANS, alpha, a_12, x_col_top, 
                              one,  y_row_right);
			PLA_Local_gemv( PLA_NO_TRANS, alpha, a_12, x_row_right,
			      one,  y_col_top);
                        PLA_Obj_implicit_local_transpose(x_row_right);
                        PLA_Obj_implicit_local_transpose(y_row_right);

		}
	}
	PLA_Obj_free( & a_11 ); PLA_Obj_free( &a_12 ); PLA_Obj_free( &a_22 );
	PLA_Obj_free( & x_row_left );  PLA_Obj_free( &x_row_right  );
	PLA_Obj_free( & x_col_top );  PLA_Obj_free( &x_col_bottom  );
	PLA_Obj_free( & y_row_left );  PLA_Obj_free( &y_row_right  );
	PLA_Obj_free( & y_col_top );   PLA_Obj_free( &y_col_bottom );
	return(PLA_SUCCESS);
}


int  PLA_Symv_lower( PLA_Obj alpha, PLA_Obj a,
		     PLA_Obj xtilde_row, PLA_Obj xtilde_col,
		     PLA_Obj one,
		     PLA_Obj ytilde_row, PLA_Obj ytilde_col)
{
	PLA_Obj a_11 = NULL, a_21 = NULL, a_22 = NULL;
	PLA_Obj x_row_left = NULL,  x_row_right  = NULL;
	PLA_Obj x_col_top = NULL,  x_col_bottom  = NULL;
	PLA_Obj y_row_left = NULL,  y_row_right  = NULL;
	PLA_Obj y_col_top  = NULL,  y_col_bottom = NULL;

	MPI_Datatype datatype;

	int     left_size, left_owner, top_size, top_owner, square_size;
	int     nb, my_proc_row, my_proc_col;
	int     rows_in_ret_buf, cols_in_ret_buf;
	PLA_Template template;

	PLA_Obj_template(a, &template);
	PLA_Temp_comm_col_rank( template, &my_proc_row );
	PLA_Temp_comm_row_rank( template, &my_proc_col );
	PLA_Temp_nb( template, &nb );
	PLA_Obj_datatype(a, &datatype);

	PLA_Obj_view_all(a, &a_22);
	PLA_Obj_view_all(xtilde_row, &x_row_right);
	PLA_Obj_view_all(xtilde_col, &x_col_bottom);
	PLA_Obj_view_all(ytilde_row, &y_row_right);
	PLA_Obj_view_all(ytilde_col, &y_col_bottom);
	while(TRUE)
	{
		PLA_Obj_split_size( a_22, PLA_SIDE_LEFT, &left_size, &left_owner);
		PLA_Obj_split_size( a_22, PLA_SIDE_TOP,  &top_size,  &top_owner );
		if( 0 == (square_size = min(left_size, top_size))) break;
		PLA_Obj_split_4(a_22, square_size, square_size, &a_11, PLA_DUMMY,
																		&a_21, &a_22     );
		PLA_Obj_horz_split_2(y_col_bottom, square_size, &y_col_top,
																  &y_col_bottom );
		PLA_Obj_horz_split_2(x_col_bottom, square_size, &x_col_top,
																  &x_col_bottom);
		PLA_Obj_vert_split_2(x_row_right,  square_size, &x_row_left, &x_row_right  );
		PLA_Obj_vert_split_2(y_row_right,  square_size, &y_row_left, &y_row_right  );

		if (my_proc_col == left_owner)
		{
                        PLA_Obj_implicit_local_transpose(x_row_left);
                        PLA_Obj_implicit_local_transpose(y_row_left);
			if( my_proc_row == top_owner)
			   PLA_Local_symv( PLA_LOWER_TRIANGULAR, alpha, 
                              a_11, x_row_left, one, y_col_top);
			PLA_Local_gemv( PLA_TRANS,  alpha, a_21, 
                                        x_col_bottom, one,  y_row_left);
			PLA_Local_gemv( PLA_NO_TRANS, alpha, a_21, x_row_left,
                                        one, y_col_bottom);
                        PLA_Obj_implicit_local_transpose(x_row_left);
                        PLA_Obj_implicit_local_transpose(y_row_left);
		}
	}
	PLA_Obj_free( & a_11 ); PLA_Obj_free( &a_21 ); PLA_Obj_free( &a_22 );
	PLA_Obj_free( & x_row_left );  PLA_Obj_free( &x_row_right  );
	PLA_Obj_free( & x_col_top );  PLA_Obj_free( &x_col_bottom  );
	PLA_Obj_free( & y_row_left );  PLA_Obj_free( &y_row_right  );
	PLA_Obj_free( & y_col_top );   PLA_Obj_free( &y_col_bottom );
	return(PLA_SUCCESS);
}

int PLA_Symv(int uplo, PLA_Obj alpha, PLA_Obj a, PLA_Obj x, PLA_Obj beta, PLA_Obj y)

{

  int owner_row, owner_col;
  PLA_Obj alpha_dup = NULL, one = NULL;
  PLA_Obj x_tilde_row = NULL, x_tilde_col = NULL;
  PLA_Obj y_tilde_row = NULL, y_tilde_col = NULL;
  MPI_Datatype datatype;
  PLA_Template template;

  PLA_Symv_enter(uplo, alpha, a, x, beta, y);

  PLA_Obj_datatype(alpha, &datatype);
  PLA_Obj_template(alpha, &template);

  PLA_Scal(beta, y);

  PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, &x_tilde_row);
  PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, &y_tilde_col);
  PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, &y_tilde_row);
  PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, &x_tilde_col);
  PLA_Copy(x, x_tilde_row);
  PLA_Copy(x, x_tilde_col);
  PLA_Obj_set_to_zero ( y_tilde_row );
  PLA_Obj_set_to_zero ( y_tilde_col );
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &one);
  PLA_Obj_set_to_one ( one);

  PLA_Obj_owner_row(alpha, &owner_row);
  PLA_Obj_owner_col(alpha, &owner_col);

  if ( owner_row != PLA_ALL_ROWS || owner_col != PLA_ALL_COLS )
  {
	  PLA_Mscalar_create_conf_to( alpha, PLA_ALL_ROWS, PLA_ALL_COLS, &alpha_dup);
	  PLA_Copy( alpha, alpha_dup );
  }

  if( PLA_UPPER_TRIANGULAR == uplo )
	  PLA_Symv_upper( alpha_dup == NULL ? alpha : alpha_dup, a,
							x_tilde_row, x_tilde_col,
							one,
							y_tilde_row, y_tilde_col);
  else
	  PLA_Symv_lower( alpha_dup == NULL ? alpha : alpha_dup, a,
							x_tilde_row, x_tilde_col,
							one,
							y_tilde_row, y_tilde_col);


  PLA_Reduce_add(y_tilde_row, PLASUM(datatype), y);
  PLA_Reduce_add(y_tilde_col, PLASUM(datatype), y);
  PLA_Obj_free( &y_tilde_row);  PLA_Obj_free( &y_tilde_col);
  PLA_Obj_free( &x_tilde_row);  PLA_Obj_free( &x_tilde_col);
  PLA_Obj_free( &one);     PLA_Obj_free( &alpha_dup);
  PLA_Symv_exit(uplo, alpha, a, x, beta, y);
  return(0);

}




