


/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/





#ifndef PLA_OBJ_H
#define PLA_OBJ_H


#ifndef NULL
#define NULL 0
#endif


typedef struct PLA_Obj_annotate_flags {
  PLA_Panel_rowcol_flag     panel_rowcol;
  PLA_Duplicate_flag        duplicate;
} PLA_Obj_annotate_flags;

typedef struct PLA_Obj_def {
  
  /* 
     Store the type of structure this is.  It can be either a 
     vector, multivector, or a matrix.
     */
  
  PLA_Obj_obj_type    obj_type;
  
  /*
     Objects can have special attributes.  Thse will depend on just
     what the object is.  For instance a matrix could be interpreted
     as a column panel.
     */

  PLA_Obj_annotate_flags  annotate_flags;
  
  /* 
     An MPI C data type.
     
     Refer to MPI documentation for acceptable values.
     Will usually be something like MPI_DOUBLE or MPI_INT.
     
     Somehow the FORTRAN wrapper to this routine will have to 
     convert the fortran datatypes, which are simply integers, to 
     C data types, which are structures.  This will probably be done
     by looking at the FORTRAN interface of the MPI and figuring out
     what they do.
     */
  MPI_Datatype  datatype;    
  
  /* 
     Buffer that will be allocated to hold data.
     */
  
  /*
     There is a problem with freeing an object.  If a submatrix object is 
     created pointing to a matrix and then the matrix is released,
     how do we make sure the submatrix is still valid for that matrix?
     Views will be an integer count of how many things point to
     this matrix.  As long as it is greater than zero the template will
     not actually be released.
     
     Master_buffer is the origional large buffer that was allocated to
     hold data.  The buffer will only be a pointer into
     this larger buffer.
     */
  int              *views;
  void             *master_buffer;  
  int               length_master_buffer_byte;
  void             *buffer;
  
  /*
     Store the origional matrix dimentions and offsets.  This will be
     only set once during creation and will not change for a submatrix
     or shifting of views.
     */
  int              master_length, master_width;
  int              master_align_row, master_align_col;

  /* 
     Store the length of the global vector and the stride within that
     vector.  Actually I am not sure if this should be a local stride
     within buffer or some kind of global stride.  Since it is not 
     implemented now, it does not matter.
     
     There is also lld, which I think is the local leading dimention 
     of the buffer if for some reason we need to stride through a matrix
     locally.  But it seems to me that after wrapping, any local stride
     would be equivalent to doing very strange things globally so I 
     don't know why we would ever use lld.
     */
  int          local_stride, local_lld;
  int          local_length, local_width;
  
  /*
     Store the global dimentions of the data structure.  There is a 
     width and a height for each of the data structures.  But for things
     like vectors the width will always be 1.  The width is the number
     of columns and the height is the number of rows.  
     
     Vectors are assumed to be column vectors.
     */

  int          global_length, global_width;

  /* 
     The template tells how the object is distributed locally.
     */
  PLA_Template  template;

  /*
     What is the offset into the template? 
     We also will store the local length since it could be used often.
     */
  int          global_align_row, global_align_col;

  /*
     A couple of obscure fields.  These are computed from the template and
     are just stored because they might be used frequently.  
     
     After distrobuting the object, there is a processor that owns the
     first row and column of the object and a processor which owns the
     last row and column of the object.  In other words one that owns the
     upper left and one that owns the lower right blocks of the object.
     
     Now the first and last block have a size to them.  They are always
     square for a matrix by definition.  For a multivector they will be the 
     width of the multivector so all we have to store is length.  
     */
  
  int          length_top_block,     owner_top_block;
  int          length_bottom_block,  owner_bottom_block;
  int          width_left_block,     owner_left_block; 
  int          width_right_block,    owner_right_block;    

/* The api field is used to enforce access protocols  */
/* when filling the array -- i.e. marking the object's */
/* buffer as "dirty", "clean", "filling" etc.          */

  void        *api;
/* The memory fields show whether a data buffer has been */
/* assigned to the object and what kind of data buffer   */
/* it is -- if it has been.                              */
  int has_data_space;
  int allocation_type;

} PLA_Obj_def;




/* ================================================================ */
/* ================================================================ */
/* ================================================================ */
/* ================================================================ */



int PLA_Obj_objtype_cast(PLA_Obj                 laobject,
                           PLA_Obj_obj_type         object_type);


/*  PLA_Obj_swap
    Simply exchanges two PLA_Obj objects.  No memory is allocated.  No memory
    is freeded.
    */
int PLA_Obj_swap (PLA_Obj *              aobj1,
		   PLA_Obj *              aobj2);

/*  PLA_Obj_init 
    Simply initialize all the fields in an PLA_Obj structure to default
    values.
    */
int PLA_Obj_init (PLA_Obj                aobj);

/*  PLA_Obj_check_type
    Check that the object type field is one of the acceptable valies. 
    Not used much now, but I would expect that someday maybe all of the
    routines would do this check before accessing fields in the structure.
    at that time maybe this should become a macro statement.
    */
int PLA_Obj_check_type (PLA_Obj          aobj);

/* PLA_Obj_free
   Frees an object created by either PLA_Mscalar_create, PLA_Vector_create,
   PLA_Mvector_create, and PLA_Matrix_create.  In other words will free any
   LA object.  
   
   Note:  Because taking subvectors and submatrixies creates pointers back
   into the origional buffers, freeing an object does not imply that all the
   buffers contained therein are freed.  A counter is kept of how many objects
   point to the allocated buffers.  Only when all objects are freeded.  
   */
int PLA_Obj_free           (PLA_Obj *       aobj);

/* PLA_Obj_objtype
   Returns what type of object this is pointing to.  One of 
   PLA_MSCALAR,
   PLA_VECTOR, 
   PLA_MVECTOR, 
   PLA_MATRIX 
   */
int PLA_Obj_objtype        (PLA_Obj           aobj, 
		             PLA_Obj_obj_type *type);

/* PLA_Obj_datatype
   What type of data is stored in this object.  This will normally
   be one of:
   MPI_DOUBLE_PRECISION or MPI_REAL
   but no checking is done and any valid MPI datatype is allowed
   */ 
int PLA_Obj_datatype       (PLA_Obj         laobject, 
		             MPI_Datatype *   datatype); 

/* PLA_Obj_global_length
   The number of rows contained within the object.   This is a global
   length, not just what exists on this processor.
   See PLA_Obj_local_length
   */
int PLA_Obj_global_length         (PLA_Obj         laobject, 
				    int *            global_length); 

/* PLA_Obj_global_width
   The number of columns contained within the object.  This is a global
   width, not just what exists on this processor.  If passed a vector it
   will always return 1.  A multivector returns the width of the multivector.
   See PLA_Obj_local_width 
   */
int PLA_Obj_global_width          (PLA_Obj         laobject, 
				    int *            global_width);

/* PLA_Obj_global_size
   PLA_Obj_global_numvecs
   These are an attempt to overcome the confusion of global_length 
   and global_width.  The problem is that when a multivector is 
   projected onto a row, the length becomes the width and the width
   becomes the length.  These two routines return the reverse, 
   unprojected sizes.  For a nonprojected object, global_size is
   equivalent to global_length and global_numvecs is the same as
   global_width.
   */
int PLA_Obj_global_size   (PLA_Obj              laobject,
                            int *                 global_size);

int PLA_Obj_global_numvecs(PLA_Obj              laobject,
                            int *                 global_numvecs);


/* PLA_Obj_local_buffer
   Returns a pointer to the buffer containing whatever data is stored locally.
   
   Note:  Many objects can point to the same data.  Only Creates allocate memory
   submatrix or subvector routines just create pointers into the origional 
   buffer.  So changing a buffer in one object could affect many objects.  
   Also you should not assume that data is always stored continuously in memory.
   If a submatrix is allocated the submatrix is just a pointer back into the
   origional big matrix, so is not continuous.
   */
int PLA_Obj_local_buffer          (PLA_Obj        laobject, 
			      void **          buffer); 

/* PLA_Obj_local_ldim
   This is the leading dimention of the array that holds the local data.  
   See PLA_Obj_local_buffer for a description why data is not alway continuous
   in memory.  
   
   Note:  This is not needed for vectors since vectors are only a one
   dimentional object.
   */
int PLA_Obj_local_ldim       (PLA_Obj        laobject,        
		              int *           local_lld); 

/*  PLA_Obj_local_stride
    The stride of the data stored in local memory.  Currently this will always
    be one and is included for future expansion only.
    */
int PLA_Obj_local_stride    (PLA_Obj        laobject,    
			      int *           local_stride); 

/* PLA_Obj_template
   Return the template associated with this object.
   
   Note: Freeing the template returned in this call will result in unpredictable
   behavior and severe errors.  DO NOT FREE A TEMPLATE OBTAINED IN THIS CALL.
   */
int PLA_Obj_template         (PLA_Obj       laobject, 
		               PLA_Template *  template); 

/* PLA_Obj_global_align_row PLA_Obj_global_align
   The global row offset from the beginning of the inducing vector.  This is set
   during the create and subvector/submatrix operations.  This values might start
   at one or zero depending on a flag set in the template. 
   */
int PLA_Obj_global_align_row (PLA_Obj       laobject,   
		               int *          global_align_row);
int PLA_Obj_global_align     (PLA_Obj       laobject,   
		               int *          global_align_row);

/* PLA_Obj_global_align_col 
   The global column offset.  Not valied for Vectors or Multivectors.  Only valid
   for matrixies.
   See PLA_Obj_global_align_row 
   */
int PLA_Obj_global_align_col (PLA_Obj       laobject,
                               int *          global_align_col); 

/* PLA_Obj_align_rowcol
   A combination of PLA_Obj_global_align_row and PLA_Obj_global_align_col
   */
int PLA_Obj_align_rowcol      (PLA_Obj       laobject,   
				int *          global_align_row,
				int *          global_align_col);

/* PLA_Obj_local_length
   Return the local length of the object  contained in laobject.
   This is not neccessarly the lenght of the buffer containing it.
   */
int PLA_Obj_local_length     (PLA_Obj       laobject,   
                               int *          local_length);

/*  PLA_Obj_local_width
    Return the local width of the object contained in laobject.  This will
    always return a 1 for vectors.  
    */
int PLA_Obj_local_width      (PLA_Obj       laobject,    
                               int *          local_width);


int PLA_Obj_local_info(PLA_Obj          obj,
                         int               *local_length,
                         int               *local_width,
                         void             **local_buffer,
                         int               *local_stride,
                         int               *local_ldim);

/*  PLA_Obj_local_all
    Return a smorgasboard of useful information.
    */
int PLA_Obj_local_all        (PLA_Obj       laobject,
			       MPI_Datatype * datatype,
			       int *          local_length,
			       int *          local_width,
			       void **        buffer,
			       int *          local_lld);


int PLA_Obj_global_info(PLA_Obj         obj,
                         int               *global_length,
                         int               *global_width,
                         int               *project,
                         int               *owner_row,
                         int               *owner_col,
                         int               *global_align_row,
                         int               *global_align_col); 


/*  PLA_Obj_global_all
    Simular to PLA_Obj_local_all
    */
int PLA_Obj_global_all       (PLA_Obj       laobject,
			       int *          global_length,
			       int *          global_width,
			       int *          global_align_row,
			       int *          global_align_col);

/*   PLA_Obj_top 
     Returns the first block of data that is contained within one row
     of processors.  It returns the length and which processor row 
     it is contained in.  
     
     This value is stored in laobject and this routine just returns it.
     A routines to compute it is PLA_Obj_first_row_block.  But it is faster to
     use this routine to retrieve it from the structure.
     
     The length will certainly be less than or equal to the block length.
     If the object is stored with a row offset, then it is possible that 
     the first block is split among two rows of processors.  This routine
     detects that and will just return the amount contained within
     the first row.  
     */
int PLA_Obj_top  (PLA_Obj       laobject,
		   int *          length_top_block,
		   int *          owner_top_block);

/* PLA_Obj_bottom
   See PLA_Obj_top
   */
int PLA_Obj_bottom  (PLA_Obj        laobject,
		      int *           length_bottom_block,
		      int *           owner_bottom_block);

/*  PLA_Obj_left
    See PLA_Obj_top
    */
int PLA_Obj_left  (PLA_Obj        laobject,
		    int *           length_first_col_block,
		    int *           owner_left_block);

/*  PLA_Obj_right
    See PLA_Obj_top
    */
int PLA_Obj_right  (PLA_Obj         laobject,
		     int *            length_last_col_block,
		     int *            owner_right_block);

int PLA_Obj_top_left      (PLA_Obj   laobject,
			    int *      length_top_left_block,
			    int *      owner_row_top_left_block,
			    int *      owner_col_top_left_block);
int PLA_Obj_top_right     (PLA_Obj   laobject,
			    int *      length_top_right_block,
			    int *      owner_row_top_right_block,
			    int *      owner_col_top_right_block);
int PLA_Obj_bottom_left   (PLA_Obj   laobject,
			    int *      length_bottom_left_block,
			    int *      owner_row_bottom_left_block,
			    int *      owner_col_bottom_left_block);
int PLA_Obj_bottom_right  (PLA_Obj   laobject,
			    int *      length_bottom_right_block,
			    int *      owner_row_bottom_right_block,
			    int *      owner_col_bottom_right_block);


int PLA_Obj_view(PLA_Obj   old_obj,
                   int         global_length,
                   int         global_width,
                   int         align_row,
                   int         align_col,
                   PLA_Obj   *new_obj); /* new view */

int PLA_Obj_view_all(PLA_Obj   old_obj,
                   PLA_Obj   *new_obj); /* new view */


/* PLA_Obj_horz_split_2
   Takes as input one object, slpits it into two objects and retures the 
   two allocated objects.  
   
   This routine is the same as calling subvector/submatrix twice.  The 
   two objects created will need to be freed sometime.
   
   NOTE:  If the pointers to upper_obj and lower_obj are not NULL on input
   these will be attempted to be freed.  Therefore you can call this routine
   in a loop without having to free the returned values since that will be
   done for you.  But it means that the first time into this routine these
   values must be set to NULL in order to avoid an error!
   
   The value in upper_length determines how much of the object to allocate to
   the upper object, the rest will be allocate into the lower object.  
   
   laobject must contain an object of length at least upper_length. 
   */
int PLA_Obj_horz_split_2     (PLA_Obj         laobject,
			       int              upper_length,
			       PLA_Obj        *upper_obj,
			       PLA_Obj        *lower_obj);

/*  PLA_Obj_vert_split_2
    This is not valid for vectors, only matrixes.
    See PLA_Obj_horz_split_2
    */
int PLA_Obj_vert_split_2     (PLA_Obj         laobject,
			       int              left_width,
			       PLA_Obj        *left_obj,
			       PLA_Obj        *right_obj);

/*  PLA_Obj_split_4 
    Same effect as calling PLA_Obj_horz_split_2 and then calling
    PLA_Obj_vert_split_2 on each of the resulting submatrixes.
    This call is only valid for matrixies. The for objects rerurned
    are all submatrixies of the origional laobject.
    
    See PLA_Obj_horz_split_2
    */
int PLA_Obj_split_4          (PLA_Obj         laobject,
			       int              upper_length,
			       int              left_width,
			       PLA_Obj        *upper_left_obj,
			       PLA_Obj        *upper_right_obj,
			       PLA_Obj        *lower_left_obj,
			       PLA_Obj        *lower_right_obj);




int PLA_Obj_view_shift(PLA_Obj       laobject,
                         int             length_top,
                         int             width_left,
                         int             width_right,
                         int             length_bottom);

int PLA_Obj_split_size(PLA_Obj      laobject,
                         int            side,
                         int           *size,
                         int           *owner);



int PLA_Obj_get_local_contents(PLA_Obj      laobject,
                                 int            transpose,
                                 int           *rows_in_buffer,
                                 int           *cols_in_buffer,
                                 void          *buffer,
                                 int            leading_dim_buffer,
                                 int            stride_buffer);

int PLA_Obj_set_local_contents(int           transpose,
                                 int            rows_in_buffer,
                                 int            cols_in_buffer,
                                 void          *buffer,
                                 int            leading_dim_buffer,
                                 int            stride_buffer,
                                 PLA_Obj       laobject);


/* PLA_Obj_extract_contents
   Copies the contents of laobject into the buffer return_buffer.
   
   The caller must allocate return_buffer and this buffer must be large
   enough to hold the contents of laobject.  The data is stored in 
   return_buffer consecutively, hence the size can be computed by calling
   PLA_Obj_local_length 
   PLA_Obj_local_width
   PLA_Obj_datatype
   Converting the datatype to a byte size with MPI_datasize
   and multiplying the resulting three nubers together.  Of course if the 
   datatype is known in advance the last step is not neccessary.  
   */
int PLA_Obj_extract_contents      (PLA_Obj         laobject,
				    int              leading_dim_ret_buf,
				    int              stride_ret_buf,
				    int  *           rows_in_ret_buf,
				    int  *           cols_in_ret_buf,
				    void *           return_buffer);

int PLA_Obj_extract_transpose_contents  (PLA_Obj    laobject,
					  int         leading_dim_ret_buf,
					  int         stride_ret_buf,
					  int  *      rows_in_ret_buf,
					  int  *      cols_in_ret_buf,
					  void *      return_buffer);


int PLA_Obj_project_onto      (PLA_Obj              laobject,
				   PLA_Panel_rowcol_flag *panel_type);
int PLA_Obj_set_orientation  (PLA_Obj              laobject,
				   PLA_Panel_rowcol_flag  panel_type);
int PLA_Obj_get_orientation  (PLA_Obj              laobject,
				   PLA_Panel_rowcol_flag  *panel_type);
int PLA_Obj_annotate_duplicate    (PLA_Obj             laobject,
                                    PLA_Duplicate_flag   *duplicate);
#endif



