


/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/






#ifndef PLA_TEMPLATE_H
#define PLA_TEMPLATE_H

/* 
   This header should be included in all source code using the PLA_Blas 
   subroutines.  It contains definitions of constants and structures 
   used in those subroutines. 
   */

#include <mpi.h>

#ifndef NULL
#define NULL 0
#endif

typedef struct PLA_Template_def {
 
  /* 
     The MPI communicator should be a 2D communicator created with
     something like MPI_Cart_create. 
     */
  MPI_Comm    comm2D;    
  
  /* 
     nb tells the blocking size of the matrix.  The matrix is 
     blocked before distributing. 
     */
  int         nb;
  
  /* 
     Communicator for row and column.  The combination will be unique
     to a processor.  It is obtained from the comm communicator during
     initialization and is just stored here for convieience. 
     Comm1D views the 2D array of processors as a single linear 1D mesh.
     */
  MPI_Comm    commrow, commcol, comm1D;

  /* 
     Topology variables.  Size of each dimension.  This is the number
     of rows and the number of columns in the matrix.  NOT the nuber of
     processors IN  a row, which would be the number of COLUMNS!
     */
  int         num_proc_row, num_proc_col;

  /*
     The indicies are the position of this node in the 2D grid as well
     as the 1D mesh.
     */
  int         myrow, mycol, me1D;

  /* 
     Finally we have this fudge factor which tells if we are numbering
     the columns of the matrix and matrix blocks starting from zero
     or one.
     0 := starting from 0;
     1 := starting from 1;
     */
  PLA_Start_num      zero_or_one;
  
  /*
     There is a problem with freeing a template.  If a matrix object is 
     created pointing to a template and then the template is released,
     how do we make sure the template is still valid for that matrix?
     Instances will be an integer count of how many things point to
     this template.  As long as it is greater than zero the template will
     not actually be released.
     */
  int              *instances;
} PLA_Template_def;


/* 
   Definitions of subroutines
   */

/* PLA_Temp_local_compute
   This is a routine used heavly my the template routines to
   determine how much of a vector is stored on each processor
   and which processor owns a certain part of a vector.
   
   This routines assumes a 1D distrobution of the vector iwth
   blocksize nb.  To find out information about matrixies, you
   have to call twice, once for row and once for column.
   
   global_length and global_offset describe where the vector
   starts and its length.
   
   processor_num is the processor, starting from zero, which you
   want information around.
   
   num_processors is the lenght of the 1D processing grid.  
   processor_num should be 0<=processor_num<num_processors.
   
   z_or_1 specifies if the global_offset starts at zero or one.
   
   local_length is how much of the vector is stored on processor
   processor_num.
   
   starting_proc is the processor number which owns the first
   element of this vector given that it starts at global_offset.
   */

int PLA_Temp_local_compute (int  global_length, 
			   int  global_offset, 
			   int  processor_num,
			   int  nb,
			   int  num_processors,
			   int  z_or_1,
			   int *local_length, 
			   int *starting_proc );

/*
   
   Create A new Template.  All multiscalar, vector, multivector, 
   and matrix creation routines need a template describing the 
   layout generating vector.  
   
   Presently there is only one generating vector that is used
   for both row and column wrapping.  The generating vector is
   distributed in COLUMN major order.  This causes a finer 
   wrapping of the rows than the columns which causes a matrix 
   to be better distributed.  
   
   Communicator should contain a valid MPI 2D communicator of 
   the type returned by MPI_Cart_create.  Row and column 
   information of the processing grid will be extraced from the
   communicator.  
   
   Blocksize will describe the blocking factor used to wrap
   the matrix or vectors.  First the object to be wrapped is
   divided up into pieces of length and (with matricies) width
   blocksize.  These blocks are then distributed over the processing
   grid.
   
   Zero_or_one is a flag that allows offsets to start at either 
   zero or one.  This flags answers the question:  Is the first
   row at 1 or 0?  Once specified all computations about the 
   MATRIX elements offsets will reference this.
   
   NOTE:  This does not apply to the PROCESSING GRID!  Processor
   rows and columns always start at 0.  Only matrix elements start
   at zero or one.  
   
   Also lengths are still normal.  A vector of length one is of
   length one wether it starts at zero or one.
   */

int PLA_Temp_create              (int            blocksize , 
				  PLA_Start_num   zero_or_one,
				  PLA_Template *  new_template);

/* PLA_Temp_free
   Free a template created by PLA_Temp_create.  Since the template
   might have been used to create some vectors and matrixies this
   routine might not actually free the memory.  Only when the last
   object that uses the template is freed will the template itself
   actually be freed.  This is done by simply keeping track of
   how many objects point to this template.
   */
int PLA_Temp_free                (PLA_Template *  old_template);

/* PLA_Temp_vector_local_length 
   Used to compute the number of rows of a vector that is stored
   on ths processor given the vector starts at global_offset
   and has length global_length.  Note that global_offset could
   start at zero or one depending on values set in the template.
   */
int PLA_Temp_vector_local_length      (PLA_Template    input_template, 
				      int            input_global_length,
				      int            input_global_offset, 
				      int *          output_local_length); 

/*  PLA_Temp_vector_distant_length
    Used like PLA_Temp_vector_local_length but allows the caller to pass in
    a processor number to compute the amount of a vector stored on that
    other processor. 
    
    Note: Processor numbers always start at zero
    */
int PLA_Temp_vector_distant_length    (PLA_Template    input_template, 
				      int            input_processor_num_1D,
				      int            input_global_length,
				      int            input_global_offset, 
				      int *          output_local_length); 

/* PLA_Temp_vector_owner_rank
   Returns the processor number of which owns the element of
   the vector at global_offset.  Global offsets can start at 
   zero or one depending on the flags set in template.  With
   vectors the processing grid is viewed as a 1D column of processors.
   */
int PLA_Temp_vector_owner_rank   (PLA_Template    input_template,
				 int            input_global_offset, 
				 int *          output_owner_rank); 

/* PLA_Temp_matrix_local_length
   Returns the number of rows of a matrix that has global_height number
   of rows and is stored starting at global_row_offset.  The offset
   can be specified at zero or one depending on flags set in the
   template.  
   Also see PLA_Temp_matrix_distant_length, PLA_Temp_matrix_local_length_width
   */
int PLA_Temp_matrix_local_length      (PLA_Template    input_template, 
				      int            input_global_height, 
				      int            input_global_row_offset,
				      int *          output_local_height); 

/*  PLA_Temp_matrix_local_width
    Like PLA_Temp_matrix_local_length, but returns the number of columns
    stored on this processor.
    Also see PLA_Temp_matrix_distant_width, PLA_Temp_matrix_local_length_width
    */
int PLA_Temp_matrix_local_width      (PLA_Template    input_template, 
				     int            input_global_width, 
				     int            input_global_col_offset, 
				     int *          output_local_width); 

/* PLA_Temp_matrix_local_length_width
   Combination of PLA_Temp_matrix_local_length and PLA_Temp_matrix_local_width  
   */
int PLA_Temp_matrix_local_length_width   (PLA_Template  input_template, 
				         int          input_global_height, 
				         int          input_global_width, 
				         int          input_global_row_offset, 
				         int          input_global_col_offset, 
				         int *        output_local_height, 
				         int *        output_local_width); 

/* PLA_Temp_matrix_distant_length
   Allows computation of how many rows of a matrix is stored on another
   processor row.  Since all processors in a single row have the same 
   number of rows of a matrix, only the processor row needs to be 
   passed.
   Also see PLA_Temp_matrix_local_length, PLA_Temp_matrix_distant_length_width
   */
int PLA_Temp_matrix_distant_length    (PLA_Template    template,
				      int            proc_row,
				      int            global_m, 
				      int            global_i,
				      int            *local_m);

/* PLA_Temp_matrix_distant_width
   Like PLA_Temp_matrix_distant_length, but computes the columns of a matrix
   owned by another column of the processing grid. 
   Also see  PLA_Temp_matrix_local_width, PLA_Temp_matrix_distant_length_width
   */
int PLA_Temp_matrix_distant_width    (PLA_Template    template, 
				     int            proc_col,
				     int            global_n, 
				     int            global_j,
				     int            *local_n);

/*  PLA_Temp_matrix_distant_length_width
    A combination of PLA_Temp_matrix_distant_length  andL_Temp_Matrix_distant_n
    Also see PLA_Temp_matrix_local_length_width 
    */
int PLA_Temp_matrix_distant_length_width   (PLA_Template    template,
					   int            proc_row,
					   int            proc_col,
					   int            global_m, 
					   int            global_n, 
					   int            global_i, 
					   int            global_j, 
					   int            *local_m, 
					   int            *local_n);

/* PLA_Temp_matrix_owner_row
   Given an offset into the inducing vector, this routine returns
   the processor row that would own that row of the matrix.  
   The offset can start at zero or one depending on the 
   flags set in the template.  
   */
int PLA_Temp_matrix_owner_row    (PLA_Template    input_template, 
				 int            input_global_row_offset,
				 int *          output_owner_row_rank); 

/* PLA_Temp_matrix_owner_col
   Like PLA_Temp_matrix_owner_row 
   */
int PLA_Temp_matrix_owner_col    (PLA_Template    input_template, 
				 int            input_global_col_offset, 
				 int *          output_owner_col_rank); 

/* PLA_Temp_matrix_owner_rowcol
   Combination of PLA_Temp_matrix_owner_row and PLA_Temp_matrix_owner_col
   */
int PLA_Temp_matrix_owner_rowcol (PLA_Template    input_template, 
				 int            input_global_row_offset, 
				 int            input_global_col_offset, 
				 int *          output_owner_row_offset, 
				 int *          output_owner_col_offset);

/* PLA_Temp_comm_col_size 
   Returns the number of processors in a row.
   */
int PLA_Temp_comm_col_size        (PLA_Template    input_template,
				 int *          output_number_of_proc_rows);

/* PLA_Temp_comm_row_size 
   Returns the number of processors in a col.
   */
int PLA_Temp_comm_row_size        (PLA_Template    input_template, 
				 int *          output_number_of_proc_cols);

/*  PLA_Temp_total_num_proc
    Returns the total number of processors in a processing grid.  This would
    also be the number of processors in the 1D grid that is defined.  
   OBSOLITE: replaced by PLA_Temp_comm_all_size
    */
int PLA_Temp_total_num_proc      (PLA_Template    input_template,
		                 int *          num_proc);

/* PLA_Temp_nb
   Returns the blocking size set in the template create routine.
   */
int PLA_Temp_nb                  (PLA_Template    input_template, 
				 int *          output_blocksize);     

/* PLA_Temp_comm_all_info
   Returns information about the 1D communicator that is used to 
   describe the processing mesh.   This 1D communicators orders the
   processing nodes in column major order.
   */
int PLA_Temp_comm_all_info(PLA_Template template,
                           MPI_Comm    *comm,
                           int         *rank,
                           int         *numnodes);

/* PLA_Temp_comm_all       replaces PLA_Temp_comm1D
   PLA_Temp_comm_all_rank  replaces PLA_Temp_me1D
   PLA_Temp_comm_all_size  replaces PLA_Temp_total_num_proc
   */
int PLA_Temp_comm_all     (PLA_Template template,
                           MPI_Comm    *comm);
int PLA_Temp_comm_all_rank(PLA_Template template,
                           int         *rank);
int PLA_Temp_comm_all_size(PLA_Template template,
                           int         *numnodes);


/* PLA_Temp_comm_row 
   Returns a row communicator that can be used in MPI calls to communicate
   within a row of the processing grid. 
   */
int PLA_Temp_comm_row             (PLA_Template    input_template, 
				 MPI_Comm *     output_row_communicator);

/* PLA_Temp_comm_row_info
   This routine returnes information about the orientation of this 
   processor within its row of processors.
   */
int PLA_Temp_comm_row_info(PLA_Template   template,
                      MPI_Comm     *comm_row,
                      int          *rank_within_row,
                      int          *numnodes_in_row);

/* PLA_Temp_comm_col
   Returns a column communicator that can be used in MPI calls to communicate
   within a column of the processing grid. 
   */
int PLA_Temp_comm_col             (PLA_Template    input_template,
				 MPI_Comm *     output_col_communicator);

/* PLA_Temp_comm_col_info
   This routine returnes information about the orientation of this 
   processor within its col of processors.
   */
int PLA_Temp_comm_col_info(PLA_Template   template,
                      MPI_Comm     *comm_col,
                      int          *rank_within_col,
                      int          *numnodes_in_col);

/* PLA_Temp_comm1D
   Returns a 1D communicator that views the processing grid in column major
   order.  This is the same order in which the inducing vector is distributed
   among the processing grid.
   OBSOLITE: replaced by PLA_Temp_comm_all
   */
int PLA_Temp_comm1D              (PLA_Template    input_template, 
				 MPI_Comm *     output_1D_communicator);

/*  PLA_Temp_comm_col_rank, PLA_Temp_comm_row_rank, and PLA_Temp_mycoords
    return the row and column of the processing grid of this processor.  
    This information is based on the row and column communicators.
    Also see PLA_Temp_comm_row PLA_Temp_comm_col and */
int PLA_Temp_comm_col_rank               (PLA_Template    input_template, 
				 int *          output_processor_row);

int PLA_Temp_comm_row_rank               (PLA_Template    input_template, 
				 int *          output_processor_col);

int PLA_Temp_mycoords            (PLA_Template    input_template, 
				 int *          output_processor_row, 
				 int *          output_processor_col);

/* PLA_Temp_me1D
   Returns the rank of this processor in based on the 1D communicator 
   returned by PLA_Temp_comm1D 
   OBSOLITE: replaced by PLA_Temp_comm_all_rank
   */
int PLA_Temp_me1D                (PLA_Template    input_template, 
				 int *          output_rank_in_1D);

/* PLA_Temp_zero_or_one
   Returns the zero_or_one flag that is stored in the template at creation time.
   This flag specifies wether offsets into the matrix or the inducing vector
   should start at zero or one.  
   Also see PLA_Temp_create
   */
int PLA_Temp_zero_or_one         (PLA_Template    input_template, 
				 PLA_Start_num * output_zero_or_one);






/* PLA_Temp_matrix/vector_first/last_row/col_block
   These routines are used to compute the values that are stored in 
   PLA_Obj structure.  There is no reason to call them 
   since the values they compute will already be stored in the laobject
   for your convieience and hence it is much quicker to just call 
   PLA_LAobj_left/right_top/bottom to retrieve it.  

   The calls are included here just for completeness and may be hidden
   in future releases.
   */
int PLA_Temp_matrix_first_row_block           (PLA_Template   template,
					      int             global_length,
					      int             global_align_row,
					      int *         owner_row,
					      int *         top_block_size);

int PLA_Temp_matrix_last_row_block            (PLA_Template   template,
					      int             global_length,
					      int             global_align_row,
					      int *         owner_row,
					      int *         bottom_block_length);

int PLA_Temp_matrix_first_col_block           (PLA_Template   template,
					      int           global_width,
					      int           global_align_col,
					      int *         owner_col,
					      int *         left_block_width);

int PLA_Temp_matrix_last_col_block            (PLA_Template   template,
					      int           global_width,
					      int           global_align_col,
					      int *         owner_col,
					      int *         right_block_width);

int PLA_Temp_matrix_first_row_first_col_block (PLA_Template   template,
					      int           global_length,
					      int           global_width,
					      int           global_align_row,
					      int           global_align_col,
					      int *         owner_row,
					      int *         owner_col,
					      int *         topleft_block_size);

int PLA_Temp_matrix_last_row_last_col_block   (PLA_Template   template,
					      int           global_length,
					      int           global_width,
					      int           global_align_row,
					      int           global_align_col,
					      int *         owner_row,
					      int *         owner_col,
					      int *         bottomright_block_size);

int PLA_Temp_matrix_first_row_last_col_block  (PLA_Template   template,
					      int           global_length,
					      int           global_width,
					      int           global_align_row,
					      int           global_align_col,
					      int *         owner_row,
					      int *         owner_col,
					      int *         topright_block_size);

int PLA_Temp_matrix_last_row_first_col_block  (PLA_Template   template,
					      int           global_length,
					      int           global_width,
					      int           global_align_row,
					      int           global_align_col,
					      int *         owner_row,
					      int *         owner_col,
					      int *         bottomleft_block_size);


int PLA_Temp_set_comm_dir(PLA_Template    template,
                          int             direction);

int PLA_Temp_get_comm_dir(PLA_Template    template,
                          int            *direction);


#endif
