/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include <limits.h>
#include <stdlib.h>
#include "mmpi.h"

/* ------------------------------------------------------------------ */

/*
#define DEBUG
*/

/* ------------------------------------------------------------------ */

#define MMPI_TAG_MIN	(0x6000)
#define MMPI_TAG_MAX	(0x7fff)

#define BITS_PER_INT    (sizeof(unsigned)*CHAR_BIT)

#define MMPI_MASK_WHO        ((unsigned long) 0x07ffffff)
#define MMPI_MASK_FLAG       ((unsigned long) 0x08000000)

/* ------------------------------------------------------------------ */

static int mmpi_comm_keyval = MPI_KEYVAL_INVALID ;

/* ------------------------------------------------------------------ */

typedef struct {

  /* MMPI_Tag_* data */

  unsigned   tmin ;
  unsigned   ltmax ;
  unsigned   gtmin ;
  unsigned   tmax ;
  unsigned   len ;
  unsigned * bits ;

  /* MMPI_Barrier data */

  MPI_Request     req ; /* Persistent request */
  int             Tag ; /* Barrier's global tag */
  unsigned        Dim ; /* Dimension of 'Who' */
  unsigned long * Who ; /* Source/destination */

  int Cycle ; /* Cycle of fan-in / fan-out */

} mmpi_comm_data ;

/* ------------------------------------------------------------------ */

static mmpi_comm_data * get_mmpi_comm_data( MPI_Comm Comm , const char name[] )
{
  mmpi_comm_data * data = (mmpi_comm_data *) NULL ;
  int flag ;
  int R ;

  if ( MPI_KEYVAL_INVALID == mmpi_comm_keyval ) {
    MMPI_Log_abort( name , "MMPI Not initialized" , MPI_ERR_UNKNOWN );
  }

  if ( MPI_SUCCESS != ( R =
       MPI_Attr_get(Comm,mmpi_comm_keyval,(void **)&data,&flag) ) ) {
    MMPI_Log_abort( name , "MPI_Attr_get failed" , R );
  }

  if ( ! flag ) {
    MMPI_Log_abort( name ,
      "MMPI not enabled for communicator" , MPI_ERR_UNKNOWN );
  }
  return data ;
}

/* ------------------------------------------------------------------ */

void MMPI_Disable( MPI_Comm comm )
{
  const char name[] = "MMPI_Disable" ;

  mmpi_comm_data * data = (mmpi_comm_data *) NULL ;
  int flag ;
  int R ;

  if ( MPI_SUCCESS != ( R = MPI_Barrier( comm ) ) ) {
    MMPI_Log_abort( name , "MPI_Barrier failed" , MPI_ERR_UNKNOWN );
  }

  if ( MPI_KEYVAL_INVALID == mmpi_comm_keyval ) {
    MMPI_Log_abort( name , "MMPI Not initialized" , MPI_ERR_UNKNOWN );
  }

  if ( MPI_SUCCESS != ( R =
       MPI_Attr_get(comm,mmpi_comm_keyval,(void **)&data,&flag) ) ) {
    MMPI_Log_abort( name , "MPI_Attr_get failed" , R );
  }

  if ( ! flag ) {
    MMPI_Log_abort( name ,
      "MMPI not enabled for communicator" , MPI_ERR_UNKNOWN );
  }

  if ( MPI_SUCCESS != ( R = MPI_Attr_delete( comm , mmpi_comm_keyval ) ) ) {
    MMPI_Log_abort( name , "MPI_Attr_delete FAILED" , R );
  }

  if ( MPI_SUCCESS != ( R =
       MPI_Attr_get(comm,mmpi_comm_keyval,(void **)&data,&flag) ) ) {
    MMPI_Log_abort( name , "MPI_Attr_get #2 failed" , R );
  }

  if ( flag ) {
    MMPI_Log_abort( name ,
      "MPI_Attr_delete was not successful" , MPI_ERR_UNKNOWN );
  }
}

/* ------------------------------------------------------------------ */

static int mmpi_comm_copy(
  MPI_Comm   oldcomm ,
  int        keyval ,
  void     * extra_state ,
  void     * attribute_val_in ,
  void     * attribute_val_out ,
  int      * flag )
{
  *((void **)attribute_val_out) = NULL ;
  *flag = 0 ;
  return MPI_SUCCESS ;
}

static int mmpi_comm_delete(
  MPI_Comm comm ,
  int      keyval ,
  void   * attribute_val ,
  void   * extra_state )
{
  const char name[] = "mmpi_comm_delete" ;
  mmpi_comm_data * data = (mmpi_comm_data *) attribute_val ;
  int R ;
  int me ;
  MPI_Status status ;

  /* Unpost the request */

  if ( MPI_SUCCESS != ( R =
          MMPI_Post_handler( data->req, NULL, MMPI_REQUEST_HANDLER_NULL ))) {
    MMPI_Log_abort( name , "Failed MMPI_Post_handler" , R );
    return R ;
  }

  /* Terminate the request */

  if ( MPI_SUCCESS != ( R = MPI_Comm_rank( comm , &me  ) ) ) {
    MMPI_Log_abort( name , "Bad MPI_Comm" , R );
    return R ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Rsend(NULL,0,MPI_INT,me,data->Tag,comm))){
    MMPI_Log_abort( name , "Failed MPI_Rsend" , R );
    return R ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Wait( &data->req, &status ) ) ) {
    MMPI_Log_abort( name , "Failed MPI_Wait" , R );
    return R ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Request_free( &data->req ) ) ) {
    MMPI_Log_abort( name , "Failed MPI_Request_free" , R );
    return R ;
  }

  free( (char *) data );

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */
/* Barrier receive handler */

static int mmpi_barrier_handler(
  void * extra_state , MPI_Request * req , MPI_Status * status )
{
  const char name[] = "MMPI_Barrier::handler" ;

  mmpi_comm_data * data  = (mmpi_comm_data *) extra_state ;
  int              Cycle = data->Cycle ;
  unsigned long  * Who   = data->Who + Cycle ;
  int R = MPI_ERR_UNKNOWN ;

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s : Cycle = %d, Who = P%d\n",name,Cycle,*Who);
  fflush(MMPI_Log_file());
#endif

  if ( data->req != *req ) {
    MMPI_Log_abort( name , "Bad request" , R );
    return R ;
  }
  if ( Cycle < 0 || data->Dim <= Cycle ) {
    MMPI_Log_abort( name , "Bad cycle in message" , R );
    return R ;
  }

  if ( ( *Who & MMPI_MASK_WHO ) != status->MPI_SOURCE ) {
    MMPI_Log_abort( name , "Bad source" , R );
    return R ;
  }
  if ( *Who & MMPI_MASK_FLAG ) {
    MMPI_Log_abort( name , "Out-of-sync message" , R );
    return R ;
  }

  *Who |= MMPI_MASK_FLAG ;

  if ( MPI_SUCCESS != ( R = MPI_Start( req ) ) ) {
    MMPI_Log_abort( name , "Failed MPI_Start" , R );
    return R ;
  }

  data->req = *req ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

int MMPI_Barrier( MPI_Comm comm )
{
  const char name[] = "MMPI_Barrier" ;
  int R ;
  int me ;
  int num ;
  int flag ;
  int cycle ;
  MPI_Status status ;
  mmpi_comm_data * data = get_mmpi_comm_data( comm , name );

  /* Perform the barrier */

  /* Query the communicator's basics */

  if ( MPI_SUCCESS != ( R = MPI_Comm_rank( comm , &me  ) ) ||
       MPI_SUCCESS != ( R = MPI_Comm_size( comm , &num ) ) ) {
    MMPI_Log_abort( name , "Bad MPI_Comm" , R ); 
    return R ;
  }

  /* Fan-in */

  for ( cycle = data->Dim - 1 ; 0 <= cycle ; cycle-- ) {
    const int upper = 1 << ( cycle + 1 );
    const int lower = 1 << cycle ;
    const int who   = me ^ lower ;

    if ( who < num ) {

      if ( lower <= who && who < upper ) {

        while ( ! ( MMPI_MASK_FLAG & data->Who[cycle] ) ) {

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s : Waiting for P%d\n",name,data->Who[cycle]);
  fflush(MMPI_Log_file());
#endif

          if ( MPI_SUCCESS != ( R = MMPI_Wait( &data->req , &status ) ) ) {
            MMPI_Log_abort( name , "Failed MMPI_Wait( receive )" , R ); 
            return R ;
          }
        }
        data->Who[cycle] &= ~MMPI_MASK_FLAG ;
      }
      else if ( lower <= me && me < upper ) {
        MPI_Request req ;

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s : Sending to P%d\n",name,who);
  fflush(MMPI_Log_file());
#endif

        if ( MPI_SUCCESS != ( R =
               MPI_Isend( &cycle, 1, MPI_INT,
                 who, data->Tag, comm , &req ) ) ) {
          MMPI_Log_abort( name , "Failed MPI_Isend" , R ); 
          return R ;
        }
        if ( MPI_SUCCESS != ( R = MMPI_Wait( &req , &status ) ) ) {
          MMPI_Log_abort( name , "Failed MMPI_Wait" , R ); 
          return R ;
        }
      }
    }
  }

  /* Fan-out */

  for ( cycle = 0 ; cycle < data->Dim ; cycle++ ) {
    const int upper = 1 << ( cycle + 1 );
    const int lower = 1 << cycle ;
    const int who   = me ^ lower ;

    if ( who < num ) {

      if ( lower <= who && who < upper ) {

        MPI_Request req ;

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s : Sending to P%d\n",name,who);
  fflush(MMPI_Log_file());
#endif

        if ( MPI_SUCCESS != ( R =
               MPI_Isend( &cycle, 1, MPI_INT,
                who, data->Tag , comm , &req ) ) ) {
          MMPI_Log_abort( name , "Failed MPI_Isend" , R ); 
          return R ;
        }
        if ( MPI_SUCCESS != ( R = MMPI_Wait( &req , &status ) ) ) {
          MMPI_Log_abort( name , "Failed MMPI_Wait" , R ); 
          return R ;
        }
      }
      else if ( lower <= me && me < upper ) {

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s : Waiting for P%d\n",name,data->Who[cycle]);
  fflush(MMPI_Log_file());
#endif

        while ( ! ( MMPI_MASK_FLAG & data->Who[cycle] ) ) {
          if ( MPI_SUCCESS != ( R = MMPI_Wait( &data->req , &status ) ) ) {
            MMPI_Log_abort( name , "Failed MMPI_Wait( receive )" , R ); 
            return R ;
          }
        }

        data->Who[cycle] &= ~MMPI_MASK_FLAG ;
      }
    }
  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */
/* Dimension of a hypercube containing 'N' processors */

static int comm_dim_cube( const int N )
  { int i ; for ( i = 0 ; ( 1 << i ) < N ; i++ ); return i ; }

/* ------------------------------------------------------------------ */

int MMPI_Enable( MPI_Comm comm )
{
  return MMPI_Enable_tag( comm , MMPI_TAG_MIN , MMPI_TAG_MAX );
}

int MMPI_Enable_tag( MPI_Comm comm , unsigned TMIN , unsigned TMAX )
{
  const char name[] = "MMPI_Enable_tag" ;
  mmpi_comm_data * data = NULL ;
  int flag ;
  int R ;

  /* If the mmpi key value does not exist create it */

  if ( mmpi_comm_keyval == MPI_KEYVAL_INVALID ) {

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s : Creating MPI key\n",name);
  fflush(MMPI_Log_file());
#endif

    if ( MPI_SUCCESS != ( R =
      MPI_Keyval_create(
        mmpi_comm_copy ,
        mmpi_comm_delete ,
        &mmpi_comm_keyval , (void *) 0 ) ) ) {
      MMPI_Log_abort( name , "Failed MPI_Keyval_create" , R );
      return R ;
    }
  }

  /* Query communicator for 'mmpi' object */

  if ( MPI_SUCCESS != ( R =
    MPI_Attr_get( comm , mmpi_comm_keyval , (void **) &data , &flag ) ) ) {
    MMPI_Log_abort( name , "Failed MPI_Attr_get" , R );
    return R ;
  }

  if ( flag ) {
    MMPI_Log_abort(
      name , "Already enabled for communicator" , MPI_ERR_UNKNOWN );
    return R ;
  }
  else {

    /* If not present, create and attach mmpi object to the communicator */

    char * ptr ;
    int me ;
    int num ;
    int dcube ;
    int cycle ;
    int ntag = 1 + TMAX - TMIN ;
    int ntag_b = ntag / BITS_PER_INT ;

    if ( ntag <= 0 ) {
      MMPI_Log_abort( name , "Bad tag range" , MPI_ERR_ARG );
    }

    if ( ntag_b % BITS_PER_INT ) ++ntag_b ;

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s : Creating handler\n",name);
  fflush(MMPI_Log_file());
#endif

    /* Query the communicator's basics */

    if ( MPI_SUCCESS != ( R = MPI_Comm_rank( comm , &me  ) ) ||
         MPI_SUCCESS != ( R = MPI_Comm_size( comm , &num ) ) ) {
      MMPI_Log_abort( name , "Bad MPI_Comm" , R );
      return R ;
    }

    dcube = comm_dim_cube( num ); /* Dimension of cube */

    /* Allocate MMPI data */

    ptr = malloc(
      sizeof(mmpi_comm_data) +
      sizeof(unsigned) * ntag_b +
      sizeof(unsigned long) * dcube );

    if ( ! ptr ) {
      MMPI_Log_abort( name , "Failed malloc" , MPI_ERR_UNKNOWN );
      return MPI_ERR_UNKNOWN ;
    }

    ptr = (char *) ( ( data = (mmpi_comm_data *) ptr ) + 1 );
    ptr = (char *) ( ( data->bits = (unsigned *) ptr ) + ntag_b );
    ptr = (char *) ( ( data->Who  = (unsigned long *) ptr ) + dcube );

    /* First global tag is reserved for the MMPI_Barrier */

    data->tmin  = TMIN ;
    data->ltmax = TMIN ;
    data->gtmin = TMAX - 1 ;
    data->tmax  = TMAX - 1 ;
    data->len   = ntag_b ;
    memset( data->bits , 0 , ntag_b * sizeof(unsigned) );

    data->req   = MPI_REQUEST_NULL ;
    data->Dim   = dcube ;
    data->Tag   = TMAX ;
    data->Cycle = -1 ;

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s : Request Tag = %d\n",name,data->Tag);
  fflush(MMPI_Log_file());
#endif

    /* Set up information */

    for ( cycle = 0 ; cycle < dcube ; cycle++ ) {

      const int upper = 1 << ( cycle + 1 );
      const int lower = 1 << cycle ;
      const int who   = me ^ lower ;

      data->Who[cycle] = MMPI_MASK_WHO ; /* An impossible number */

      if ( who < num )
          if (  ( lower <= who && who < upper )   /* Fan-in  receive */
             || ( lower <= me  &&  me < upper ) ) /* Fan-out receive */
              data->Who[ cycle ] = who ;
    }

    /* Post barrier receive */

    if ( MPI_SUCCESS != ( R =
            MPI_Recv_init( &data->Cycle , 1 , MPI_INT ,
              MPI_ANY_SOURCE , data->Tag , comm , &data->req ) ) ) {
      MMPI_Log_abort( name , "Failed MPI_Recv_init" , R );
      return R ;
    }

    if ( MPI_SUCCESS != ( R =
            MMPI_Post_handler( data->req , (void *) data ,
              (MMPI_Request_handler *) mmpi_barrier_handler ) ) ) {
      MMPI_Log_abort( name , "Failed MMPI_Post_handler" , R );
      return R ;
    }

    if ( MPI_SUCCESS != ( R = MPI_Start( &data->req ) ) ) {
      MMPI_Log_abort( name , "Failed MPI_Start" , R );
      return R ;
    }

    if ( MPI_SUCCESS != ( R =
      MPI_Attr_put( comm , mmpi_comm_keyval , data ) )) {
      MMPI_Log_abort( name , "Failed MPI_Attr_put" , R );
      return R ;
    }
  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

static void bit_flip( unsigned * b , int n )
{
  b[n / BITS_PER_INT] ^= ((unsigned)01) << (n % BITS_PER_INT);
}

static int bit_value( unsigned * b , int n )
{
  return (b[n / BITS_PER_INT] >> (n % BITS_PER_INT)) & 01 ;
}

/* ------------------------------------------------------------------ */

int MMPI_Tag_verify( MPI_Comm Comm )
{
  static char name[] = "MMPI_Tag_verify" ;
  int ngtable ;
  int ngtag ;
  unsigned * gtable = NULL ;
  mmpi_comm_data * data = get_mmpi_comm_data( Comm , name );

  ngtag   = 1 + data->tmax - data->gtmin ;
  ngtable = ngtag / BITS_PER_INT ;
  if ( ngtag % BITS_PER_INT ) ++ngtable ;

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s(ngtag=%d,ngtable=%d)\n",name,ngtag,ngtable);
#endif

  gtable = malloc( sizeof(unsigned) * ( ngtable + 2 ) );

  gtable[0] = data->gtmin ;
  gtable[1] = data->tmax ;
  memcpy( gtable + 2 , data->bits + data->len - ngtable,
    ngtable * sizeof(unsigned) );

  /* Verify against P0 */

  MPI_Bcast( gtable , ngtable + 2 , MPI_UNSIGNED , 0 , Comm );

  if ( gtable[0] != data->gtmin ||
       gtable[1] != data->tmax ||
       memcmp( gtable + 2 , data->bits + data->len - ngtable ,
         ngtable * sizeof(unsigned) ) ) {
    MMPI_Log_message( name , "Communicator's tag table is out-of-sync" );
    return MPI_ERR_COMM ;
  }

  free( gtable );
  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

int MMPI_Tag_get_global( MPI_Comm Comm , int * Tag )
{
  const char name[] = "MMPI_Tag_get_global" ;
  int T = -1 ;
  int i ;
  mmpi_comm_data * data = get_mmpi_comm_data( Comm , name );

  for ( i = 1 ;
    i <= data->len && data->bits[data->len - i] == ~((unsigned)0) ; i++ );

  if ( i <= data->len ) {

    T = data->tmax - ( i - 1 ) * BITS_PER_INT ;

    {
      unsigned bits = data->bits[ data->len - i ] ;
      unsigned mask = 01 << ( BITS_PER_INT - 1 );
      for ( ; mask && ( mask & bits ) ; T-- , mask >>= 1 );
    }

    if ( data->ltmax < T ) {
      if ( T < data->gtmin ) data->gtmin = T ;
      bit_flip( data->bits , data->len * BITS_PER_INT - (1 + data->tmax - T) );
      *Tag = T ;
      return MPI_SUCCESS ;
    }
  }

  MMPI_Log_abort( name , "No available tags" , MPI_ERR_UNKNOWN );

  return MPI_ERR_UNKNOWN ;
}

int MMPI_Tag_rel_global( MPI_Comm Comm , int * Tag )
{
  static char name[] = "MMPI_Tag_rel_global" ;
  int T = *Tag ;
  mmpi_comm_data * data = get_mmpi_comm_data( Comm , name );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s(,Tag = %d = %d)\n",name,*Tag,T);
#endif

  if ( data->gtmin <= T && T <= data->tmax ) {
    T = data->len * BITS_PER_INT - ( 1 + data->tmax - T );
    if ( bit_value( data->bits , T ) ) {
      bit_flip( data->bits , T );
      *Tag = -1 ;
      return MPI_SUCCESS ;
    }
  }

  MMPI_Log_message( name , "Warning: released tag was not reserved" );

  return MPI_ERR_UNKNOWN ;
}

/* ------------------------------------------------------------------ */

int MMPI_Tag_get_local( MPI_Comm Comm , int * Tag )
{
  const char name[] = "MMPI_Tag_get_local" ;
  int T = -1 ;
  int i ;
  mmpi_comm_data * data = get_mmpi_comm_data( Comm , name );

  for ( i = 0 ; i < data->len && data->bits[i] == ~((unsigned)0) ; i++ );

  if ( i < data->len ) {
    T = data->tmin + i * BITS_PER_INT ;
    {
      unsigned bits = data->bits[i] ;
      unsigned mask = 01 ;
      for ( ; mask && ( mask & bits ) ; T++ , mask <<= 1 );
    }

    if ( T < data->gtmin ) {
      if ( data->ltmax < T ) data->ltmax = T ;
      bit_flip( data->bits , T - data->tmin );
      *Tag = T ;
      return MPI_SUCCESS ;
    }
  }

  MMPI_Log_abort( name , "No available tags" , MPI_ERR_UNKNOWN );

  return MPI_ERR_UNKNOWN ;
}

int MMPI_Tag_rel_local( MPI_Comm Comm , int * Tag )
{
  static char name[] = "MMPI_Tag_rel_local" ;
  int T = *Tag ;
  mmpi_comm_data * data = get_mmpi_comm_data( Comm , name );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s(,Tag = %d = %d)\n",name,*Tag,T);
#endif

  if ( data->tmin <= T && T <= data->ltmax ) {

    if ( bit_value( data->bits , T -= data->tmin ) ) {
      bit_flip( data->bits , T );
      *Tag = -1 ;
      return MPI_SUCCESS ;
    }
  }

  MMPI_Log_message( name , "Warning: released tag was not reserved" );

  return MPI_ERR_UNKNOWN ;
}

/* ------------------------------------------------------------------ */

