#include "PLA.h"
#include "mmpi.h"

/******************************************************************************/
/******************************************************************************/
/*   END OF HEADERS ***********************************************************/
/*   START OF CODE  ***********************************************************/
/******************************************************************************/
/******************************************************************************/

static int PLA_SEQUENTIAL_CHECK_FLAG = PLA_NO_CHECK_SEQUENTIAL;
static int PLA_INSTRUMENT_FLAG = PLA_NO_INSTRUMENT;
static int PLA_CHECK_PARAMETER_FLAG = PLA_NO_CHECK_PARAMETER;


/**************** START OF CORE ROUTINES **************************************/
void pla_Log_abort(
  const char * const who ,   /* function */
  const char * const file,
  const int          line,
  const char * const msg ,
  int error,
  int original_error )
{
   PLA_Stack_frame_obj tos;
   char next_message[120];
   char String_buffer[120];
   char Error_coded_message[80],Original_coded_message[80];

   pla_errorcode_string(error, Error_coded_message);
   pla_errorcode_string(original_error, Original_coded_message);
   sprintf(String_buffer, "%s in file %s on line %5d incurred %s a FATAL error : %s",
                          who, file, line, Error_coded_message, Original_coded_message);
   PLA_Stack_get_tos(&tos);            /* nec. the original error.       */
   while(tos != NULL)
   {
      PLA_Stack_pop_current(next_message);
      MMPI_Log_message("Stack Frame : ", next_message);
      PLA_Stack_get_tos(&tos);
   }
/* MMPI_Log_abort shuts stuff down.  It doesn't seem unreasonable to see  */
/* the actual fatal error message at the top of the stack and then at the */
/* very end.                                                              */
   MMPI_Log_abort(String_buffer, msg, error);
}

int pla_Log_error(
   const char * const who ,
   const char * const file,
   const int          line,
   const char * const msg,
   int          this_error,
   int          original_error
   )
{
   char String_buffer[120];
   char Error_coded_message[80], Original_coded_message[80];
   int internal_error = 0;

   pla_errorcode_string(this_error, Error_coded_message);
   pla_errorcode_string(original_error, Original_coded_message);
   sprintf(String_buffer, "%s in file %s on line %5d incurred %s : %s",
                          who, file, line, Original_coded_message,
                          Error_coded_message);
   MMPI_Log_message(String_buffer, msg);
   return ( PLA_RETURN_LOG_ERROR );
}

int pla_Log_message(
   const char * const who ,
   const char * const file,
   const int          line,
   const char * const msg
   )
{
   char String_buffer[80];
   sprintf(String_buffer, "%s in file %s on line %5d", who, file, line);
   MMPI_Log_message(String_buffer, msg);
   return ( PLA_RETURN_LOG_MESSAGE );
}
/**************** END OF CORE ROUTINES ****************************************/

/*************  START "STANDARD" ERROR TESTS **********************************/

int pla_Obj_object_valid(
   PLA_Obj obj,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error)
{
   char string_buffer[80];
   int  error = 0, old_error = 0;

   if(error = pla_Obj_datatype_valid(obj, file, line, message, original_error))
      pla_Log_abort("PLA_Object_valid", file, line, message, error, original_error);
   else if(error = pla_Obj_objtype_valid(obj, file, line, message, original_error))
      pla_Log_abort("PLA_Object_valid", file, line, message, error, original_error);
/* More checks to be added of course */
   /* else if */
      return PLA_SUCCESS;
}

int pla_Obj_objtype_valid(
   PLA_Obj object, 
   const char * const file,
   const int          line,
   const char * const message,
   int original_error )
{
   int  objtype;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_objtype(object, &objtype);
   if( (objtype != PLA_MSCALAR) && (objtype != PLA_VECTOR) &&
       (objtype != PLA_MVECTOR) && (objtype != PLA_MATRIX) &&
       (objtype != PLA_PVECTOR) && (objtype != PLA_PMVECTOR))
   {
      sprintf(string_buffer, "Unknown object type for %s", message);
      pla_Log_abort("PLA_Obj_objtype_valid", file, line, string_buffer, PLA_INVALID_OBJTYPE, original_error);
      return PLA_INVALID_DATATYPE;
   }
   else
      return PLA_SUCCESS;
}





int pla_Obj_objtype_equal( 
	PLA_Obj obj, 
	int object_type, 
    const char * const file,
    const int          line,
	const char * const message,
    int original_error )
{
   int object_type_of_target;
   char string_buffer[80], string_source[15], string_target[15];
   int  error = 0;

   PLA_Obj_objtype(obj, &object_type_of_target);
   if(object_type_of_target != object_type)
   {
      pla_code_string_objtype(object_type, string_source);
      pla_code_string_objtype(object_type_of_target, string_target);
      sprintf(string_buffer, "Object type of %s, %s, did not match %s",
              message, string_target, string_source);
	  pla_Log_error("PLA_Obj_objtype_equal", file, line, string_buffer, PLA_OBJTYPE_MISMATCH, original_error);

      return(PLA_OBJTYPE_MISMATCH);
   }
   else
      return PLA_SUCCESS;
}


int pla_Obj_datatype_valid(
   PLA_Obj object, 
   const char * const file,
   const int          line,
   const char * const message,
   int original_error )
{
   MPI_Datatype datatype;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_datatype(object, &datatype);
   if( (datatype != MPI_DOUBLE) && (datatype != MPI_FLOAT) &&
       (datatype != MPI_INT)   
     /* && (datatype != MPI_COMPLEX) && (datatype != MPI_DOUBLE_COMPLEX) */ 
       )
   {
      sprintf(string_buffer, "Unknown datatype for %s", message);
      pla_Log_abort("PLA_Obj_datatype_valid", file, line, string_buffer, PLA_INVALID_DATATYPE, original_error);
      return PLA_INVALID_DATATYPE;
   }
   else
      return PLA_SUCCESS;
}


int pla_Obj_datatype_equal( 
   PLA_Obj object, 
   MPI_Datatype datatype, 
   const char * const file,
   const int          line,
   const char * const message,
   int original_error )
{
   MPI_Datatype datatype_target;
   char string_buffer[80], type_target[12], type_input[12];
   int  error = 0;

   PLA_Obj_datatype(object, &datatype_target);
   if(datatype != datatype_target)
   {
      pla_code_string_datatype(datatype, type_input);
      pla_code_string_datatype(datatype_target, type_target);
      sprintf(string_buffer, "Datatype mismatch : %s type %s instead of %s",
                             message, type_target, type_input);
      pla_Log_error("PLA_Obj_datatype_equal", file, line, string_buffer, PLA_DATATYPE_MISMATCH, original_error);
   }
   else
      return PLA_SUCCESS;
}

int pla_Obj_template_equal( 
   PLA_Obj object, 
   PLA_Template template, 
   const char * const file,
   const int          line,
   const char * const message,
   int original_error )
{
   PLA_Template template_target;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_template(object, &template_target);
   if(template != template_target) /* only one -- ptr comparison is all we need */
   {
      sprintf(string_buffer, "Object %s has non-matching template", message);
      pla_Log_error("PLA_Obj_template_equal", file, line, string_buffer, PLA_TEMPLATE_MISMATCH, original_error);
      return PLA_TEMPLATE_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}


int pla_Obj_global_length_equal ( 
	PLA_Obj object, 
	int size, 
    const char * const file,
    const int          line,	
	const char * const message,
    int original_error )
{
   int object_length;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_global_length(object, &object_length);
   if(object_length != size)
   {

      sprintf(string_buffer,"Length mismcatch: %s of length %d should be %d",
                              message, object_length, size);
      pla_Log_error("PLA_Obj_global_length_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}


int pla_Obj_global_width_equal ( 
   PLA_Obj object, 
   int size, 
   const char * const file,
   const int          line,
   const char * const message,
   int original_error )
{
   int object_width;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_global_width(object, &object_width);
   if(object_width != size)
   {

      sprintf(string_buffer,"Width mismcatch: %s of width %d should be %d",
                              message, object_width, size);
      pla_Log_error("PLA_Obj_global_width_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}



int pla_Obj_local_length_equal ( 
   PLA_Obj object, 
   int size, 
   const char * const file,
   const int          line,   
   const char * const message,
   int original_error )
{
   int object_length;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_local_length(object, &object_length);
   if(object_length != size)
   {

      sprintf(string_buffer,"Length mismcatch: %s of length %d should be %d",
                              message, object_length, size);
      pla_Log_error("PLA_Obj_local_length_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}


int pla_Obj_local_width_equal ( 
   PLA_Obj object, 
   int size, 
   const char * const file,
   const int          line,   
   const char * const message,
   int original_error )
{
   int object_width;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_local_width(object, &object_width);
   if(object_width != size)
   {

      sprintf(string_buffer,"Width mismatch: %s of width %d should be %d",
                              message, object_width, size);
      pla_Log_error("PLA_Obj_local_width_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
   }
   else
      return PLA_SUCCESS;
}



int pla_Obj_global_length_length_equal ( 
   PLA_Obj obj1, 
   PLA_Obj obj2, 
   const char * const file,
   const int          line,
   const char * const message,
   int original_error )
{
   int obj1_length, obj2_length;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_global_length(obj1, &obj1_length);
   PLA_Obj_global_length(obj2, &obj2_length);
   if(obj1_length != obj2_length)
   {
      sprintf(string_buffer, "Length vs. Length mismatch : %s %d vs. %d",
                 message, obj1_length, obj2_length);
      pla_Log_error("PLA_Obj_global_length_length_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}


int pla_Obj_global_length_width_equal ( 
	PLA_Obj obj1, 
	PLA_Obj obj2, 
    const char * const file,
    const int          line,	
	const char * const message,
    int original_error )
{
   int obj1_length, obj2_width;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_global_length(obj1, &obj1_length);
   PLA_Obj_global_width(obj2, &obj2_width);
   if(obj1_length != obj2_width)
   {
      sprintf(string_buffer, "Length vs. Width mismatch : %s %d vs. %d",
                 message, obj1_length, obj2_width);
      pla_Log_error("PLA_Obj_global_length_width_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}


int pla_Obj_global_width_length_equal ( 
	PLA_Obj obj1, 
	PLA_Obj obj2, 
    const char * const file,
    const int          line,	
	const char * const message,
    int original_error )
{
   int obj1_width, obj2_length;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_global_width(obj1, &obj1_width);
   PLA_Obj_global_length(obj2, &obj2_length);
   if(obj1_width != obj2_length)
   {
      sprintf(string_buffer, "Width vs. Length mismatch : %s %d vs. %d",
                 message, obj1_width, obj2_length);
      pla_Log_error("PLA_Obj_global_width_length_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}

int pla_Obj_global_width_width_equal ( 
	PLA_Obj obj1, 
	PLA_Obj obj2, 
    const char * const file,
    const int          line,	
	const char * const message,
    int original_error )
{
   int obj1_width, obj2_width;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_global_width(obj1, &obj1_width);
   PLA_Obj_global_width(obj2, &obj2_width);
   if(obj1_width != obj2_width)
   {
      sprintf(string_buffer, "Width vs. Width mismatch : %s %d vs. %d",
                 message, obj1_width, obj2_width);
      pla_Log_error("PLA_Obj_global_width_width_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}





int pla_Obj_local_length_length_equal ( 
	PLA_Obj obj1, 
	PLA_Obj obj2, 
    const char * const file,
    const int          line,	
	const char * const message,
    int original_error )
{
   int obj1_length, obj2_length;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_local_length(obj1, &obj1_length);
   PLA_Obj_local_length(obj2, &obj2_length);
   if(obj1_length != obj2_length)
   {
      sprintf(string_buffer, "Length vs. Length mismatch : %s %d vs. %d",
                 message, obj1_length, obj2_length);
      pla_Log_error("PLA_Obj_local_length_length_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}

int pla_Obj_local_length_width_equal ( 
	PLA_Obj obj1, 
	PLA_Obj obj2, 
    const char * const file,
    const int          line,	
	const char * const message,
    int original_error )
{
   int obj1_length, obj2_width;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_local_length(obj1, &obj1_length);
   PLA_Obj_local_width(obj2, &obj2_width);
   if(obj1_length != obj2_width)
   {
      sprintf(string_buffer, "Length vs. Width mismatch : %s %d vs. %d",
                 message, obj1_length, obj2_width);
      pla_Log_error("PLA_Obj_local_length_width_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}

int pla_Obj_local_width_length_equal ( 
	PLA_Obj obj1, 
	PLA_Obj obj2, 
    const char * const file,
    const int          line,	
	const char * const message,
    int original_error )
{
   int obj1_width, obj2_length;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_local_width(obj1, &obj1_width);
   PLA_Obj_local_length(obj2, &obj2_length);
   if(obj1_width != obj2_length)
   {
      sprintf(string_buffer, "Width vs. Length mismatch : %s %d vs. %d",
                 message, obj1_width, obj2_length);
      pla_Log_error("PLA_Obj_local_width_length_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}

int pla_Obj_local_width_width_equal ( 
	PLA_Obj obj1, 
	PLA_Obj obj2, 
    const char * const file,
    const int          line,	
	const char * const message,
    int original_error )
{
   int obj1_width, obj2_width;
   char string_buffer[80];
   int  error = 0;

   PLA_Obj_local_width(obj1, &obj1_width);
   PLA_Obj_local_width(obj2, &obj2_width);
   if(obj1_width != obj2_width)
   {
      sprintf(string_buffer, "Width vs. Width mismatch : %s %d vs. %d",
                 message, obj1_width, obj2_width);
      pla_Log_error("PLA_Obj_local_width_width_equal", file, line, string_buffer, PLA_DIMENSION_MISMATCH, original_error);
      return PLA_DIMENSION_MISMATCH;
   }
   else
      return PLA_SUCCESS;
}
/*************  END "STANDARD" ERROR TESTS ************************************/


/************* START ENCODED --> STRING CONVERSIONS  **************************/
int pla_code_string_objtype(int Object_type, char *object_name)
{
   switch(Object_type) {
      case PLA_MATRIX : strcpy(object_name, "PLA_MATRIX");
                        break;
      case PLA_VECTOR : strcpy(object_name, "PLA_VECTOR");
                        break;
      case PLA_MVECTOR : strcpy(object_name, "PLA_MVECTOR");
                        break;
      case PLA_MSCALAR : strcpy(object_name, "PLA_MSCALAR");
                        break;
      case PLA_PVECTOR : strcpy(object_name, "PLA_PVECTOR");
                        break;
      case PLA_PMVECTOR : strcpy(object_name, "PLA_PMVECTOR");
                        break;
      default : strcpy(object_name, "UNDEFINED");
   }
}

int pla_code_string_datatype(MPI_Datatype Datatype, char *data_name)
{
   char object_name[40];

   if(Datatype == MPI_FLOAT) strcpy(object_name, "MPI_FLOAT");
   else if (Datatype ==   MPI_DOUBLE ) strcpy(object_name, "MPI_DOUBLE");
   else if (Datatype ==   MPI_INT ) strcpy(object_name, "MPI_INT");
/* These were dying -- but James may have fixed them by now */
/*   else if (Datatype ==   MPI_COMPLEX ) strcpy(object_name, "MPI_COMPLEX");
   else if (Datatype ==   MPI_DOUBLE_COMPLEX ) strcpy(object_name, "MPI_DOUBLE_COMPLEX"); */
   else strcpy(object_name, "UNDEFINED");
}

void pla_errorcode_string(int error_code, char *error_code_string)
{
/* Dummy stub at this point */
   if(error_code < 0) strcpy(error_code_string, "Some warning type deal");
   else if(error_code > 0) strcpy (error_code_string, "Some serious type deal");
   else /* error_code == 0 */ strcpy(error_code_string, "Not an error deal");
   return;
}
/************* END ENCODED --> STRING CONVERSIONS  ****************************/

/**************** START FLAG INQUIRY ROUTINES *********************************/

int   PLA_Get_check_sequential (int *check_sequential )
{
   *check_sequential = PLA_SEQUENTIAL_CHECK_FLAG;
   return PLA_SUCCESS;
}

int   PLA_Set_check_sequential (int sequential_set )
{
   PLA_SEQUENTIAL_CHECK_FLAG = sequential_set;
   return PLA_SUCCESS;
}

int   PLA_Get_instrument (int *instrument )
{
   *instrument = PLA_INSTRUMENT_FLAG;
   return PLA_SUCCESS;
}

int   PLA_Set_instrument (int instrument_set )
{
   PLA_INSTRUMENT_FLAG = instrument_set;
   return PLA_SUCCESS;
}

int  PLA_Get_check_parameters ( int *check_parameters )
{
   *check_parameters = PLA_CHECK_PARAMETER_FLAG; /* 0 or 1, yes or no */
}

int  PLA_Set_check_parameters ( int parameters_set )
{
   PLA_CHECK_PARAMETER_FLAG = parameters_set; /* 0 or 1, yes or no */
}

/**************** END FLAG INQUIRY ROUTINES ***********************************/

/**************** START TEMPORARY COPY ROUTINE ********************************/
/* Copy a matrix to an mscalar */
/* This will be replaced by Philip's PLA_Copy() as soon as it is available */

int PLA_Copy_temp(PLA_Obj the_matrix, PLA_Obj the_mscalar)
{
   int           length, width, owner_row, owner_col, zero_or_one;
   int           mscalar_lld;
   PLA_Obj       one = NULL;
   MPI_Datatype  datatype;
   void         *one_buffer, *mscalar_buffer;
   PLA_Template  template;

   PLA_API_begin();
   PLA_Obj_API_open(the_matrix);
   PLA_Obj_API_open(the_mscalar);
   PLA_Obj_global_length(the_matrix, &length); /* if these were inlined - */
   PLA_Obj_global_width (the_matrix, &width);  /* they would have been tested*/
   PLA_Obj_local_buffer(the_mscalar, &mscalar_buffer);
   PLA_Obj_local_ldim(the_mscalar, &mscalar_lld);
   PLA_Obj_owner_row(the_mscalar, &owner_row);
   PLA_Obj_owner_col(the_mscalar, &owner_col);
   PLA_Obj_datatype(the_matrix, &datatype); /* assume? they are the same? */
   PLA_Obj_template(the_matrix, &template);

   PLA_Temp_zero_or_one(template, &zero_or_one);

   PLA_Mscalar_create(datatype, owner_row, owner_col, length, width,
                      template, &one);
   PLA_Obj_set_to_one(one);
   PLA_Obj_local_buffer(one, &one_buffer);

   PLA_API_axpy_global_to_matrix(length, width, (void*)one_buffer,
                                 the_matrix, zero_or_one, zero_or_one,
                                 (void*)mscalar_buffer, mscalar_lld);
   PLA_Obj_API_close(the_mscalar);
   PLA_Obj_API_close(the_matrix);
   PLA_API_end();
   PLA_Obj_free( &one );
}

/**************** END TEMPORARY COPY ROUTINE **********************************/


/**************** START STACK STUFF  ******************************************/
int PLA_Stack_get_tos( PLA_Stack_frame_obj *top_of_stack)
{
   *top_of_stack = PLA_TOP_OF_MESSAGE_STACK; /* global */
   return 0;
}

int PLA_Stack_set_tos( PLA_Stack_frame_obj  top_of_stack)
{
   PLA_TOP_OF_MESSAGE_STACK = top_of_stack;
   return 0;
}

/* Maybe PLA_Push_current should have two parameters -- stack name and message*/

int PLA_Stack_push_current(char *Stack_string)
{
   PLA_Stack_frame_obj tos, new_frame;
   int string_size;

   PLA_Stack_get_tos(&tos);
   new_frame = (PLA_Stack_frame_obj) pla_malloc(sizeof(PLA_Stack_node));
   string_size = strlen(Stack_string)+1;
   new_frame->Message = (char*) pla_malloc(sizeof(char)*string_size);
   strcpy(new_frame->Message, Stack_string);
   new_frame->previous_frame = tos;
   new_frame->next_frame = NULL;
   if(tos != NULL)
      tos->next_frame = new_frame;
   PLA_Stack_set_tos(new_frame);
   return 0;
}

int PLA_Stack_pop_current(char *Stack_string)
{
   PLA_Stack_frame_obj tos, new_tos;
   int string_size;

   PLA_Stack_get_tos(&tos);
   strcpy(Stack_string, tos->Message);
   pla_free(tos->Message);
   new_tos = tos->previous_frame;
   PLA_Stack_set_tos(new_tos);
   pla_free(tos);
   if(new_tos != NULL)
   	new_tos->next_frame = NULL;
   return 0;
}

/**************** END OF STACK STUFF *****************************************/

