#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

#define MISC_DEBUG_NO

/*********************************************************
* Miscellenous common linear algebra functions. 
*
* Math functions:
*
* void PRISM_min( PLA_Obj X, PLA_Obj alpha )
*      X: Matrix, Mvector
*      alpha: Mscalar
* void PRISM_max( PLA_Obj X, PLA_Obj alpha )
*      X: Matrix, Mvector
*      alpha: Mscalar
* void PRISM_sum( PLA_Obj X, PLA_Obj alpha )
*      X: Matrix, Mvector
*      alpha: Mscalar
* void PRISM_abs( PLA_Obj X, PLA_Obj Y )
*      X: Any PLAPACK object
*      Y: conf_to X
* void PRISM_asum( PLA_Obj X, PLA_Obj Y )
*      X: Matrix
*      Y: Mvector
* void PRISM_norm1( PLA_Obj X, PLA_Obj alpha )
*      X: Matrix
*      alpha: Mscalar
* void PRISM_trace( PLA_Obj X, PLA_Obj alpha )
*      X: Matrix
*      alpha: Mscalar
* void PRISM_trans( PLA_Obj X, PLA_Obj T )
*      X: Matrix, Mvector
*      T: Matrix, Mvector
* void PRISM_get_diag( PLA_Obj X, PLA_Obj D )
*      X: Matrix
*      D: Mvector (column oriented)
* void PRISM_set_diag( PLA_Obj X, PLA_Obj D );
*      X: Matrix
*      D: Mvector (column oriented)
* void PRISM_set_to_ident( PLA_Obj X )
*      X: Matrix
* void PRISM_add_to_diag( PLA_Obj alpha, PLA_Obj X );
*      alpha: Mscalar
*      X: Matrix
* void PRISM_sym( PLA_Obj X )
*      X: Matrix
*
* Utilities:
* double PRISM_mscalar     ( PLA_Obj alpha )
* void PRISM_print         ( PLA_Obj X )
* void PRISM_long_print    ( PLA_Obj X )
*
* Notes:
* 1. MPI_Datatype is assumed to be MPI_DOUBLE.
*********************************************************/

/* Minimum component of a PLAPACK object. */
void PRISM_min(PLA_Obj X, PLA_Obj alpha) 
{
    int i, length, width;
    void *buffer;
    double *x, keep;
    PLA_Obj temp=NULL, view=NULL;

    /* Temporary multiscalar from PLA_Reduce */
    PLA_Mscalar_create_conf_to( alpha, PLA_ALL_ROWS, PLA_ALL_COLS, &temp ); 

    /* Initialize temp to first element in vector */
    PLA_Obj_view(X, 1, 1, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &view);
    PLA_Copy( view, temp );
    PLA_Obj_free( &view );

    PLA_Obj_local_length  ( X, &length );
    PLA_Obj_local_width   ( X, &width );
    PLA_Obj_local_buffer  ( X, &buffer );
    x = (double *)buffer;

    if (length*width > 0) {
       keep = x[0];
       for (i=1; i<length*width; i++) 
           keep = MIN( keep, x[i] );
       PLA_Obj_set( temp, MPI_DOUBLE, &keep );
    } 
    PLA_Reduce( temp, MPI_MIN, alpha );
    PLA_Obj_free( &temp );

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_min\n");
    PRISM_print(X);
    PRISM_print(alpha);
#endif
}

/* Maximum component of a PLAPACK object. */
void PRISM_max(PLA_Obj X, PLA_Obj alpha) 
{
    int i, length, width;
    void *buffer;
    double *x, keep;
    PLA_Obj temp=NULL, view=NULL;

    /* Temporary multiscalar from PLA_Reduce */
    PLA_Mscalar_create_conf_to(alpha, PLA_ALL_ROWS, PLA_ALL_COLS, &temp); 

    /* Initialize temp to first element in vector */
    PLA_Obj_view(X, 1, 1, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &view);
    PLA_Copy( view, temp );
    PLA_Obj_free( &view );

    /* Set temp to local max */
    PLA_Obj_local_length  ( X, &length );
    PLA_Obj_local_width   ( X, &width );
    PLA_Obj_local_buffer  ( X, &buffer );    
    x = (double *)buffer;
 
    if (length*width > 0) {
       keep = x[0];
       for (i=1; i<length*width; i++) 
           keep = MAX( keep, x[i] );
       PLA_Obj_set( temp, MPI_DOUBLE, &keep );
    } 
    PLA_Reduce( temp, MPI_MAX, alpha );
    PLA_Obj_free( &temp );

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_max\n");
    PRISM_print(X);
    PRISM_print(alpha);
#endif
}

/* Sum the components of a PLAPACK object. */
void PRISM_sum(PLA_Obj X, PLA_Obj alpha) 
{
    int i, length, width;
    void *buffer;
    double *x, keep;
    PLA_Obj temp=NULL;

    /* Temporary multiscalar from PLA_Reduce */
    PLA_Mscalar_create_conf_to(alpha, PLA_ALL_ROWS, PLA_ALL_COLS, &temp); 

    PLA_Obj_local_length ( X, &length );
    PLA_Obj_local_width  ( X, &width );
    PLA_Obj_local_buffer ( X, &buffer );
    x = (double *)buffer;

    keep = 0.;
    for (i=0; i<length*width; i++) 
        keep += x[i];
    PLA_Obj_set( temp, MPI_DOUBLE, &keep );
    PLA_Reduce( temp, MPI_SUM, alpha );
    PLA_Obj_free( &temp );

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_sum\n");
    PRISM_print(X);
    PRISM_print(alpha);
#endif
}


/* Absolute value of a PLAPACK object. */
/* Assumes X and Y are conf_to */
void PRISM_abs(PLA_Obj X, PLA_Obj Y) 
{
    int i, length, width;
    void *buffer;
    double *y;

    PLA_Local_copy( X, Y );

    PLA_Obj_local_length  ( Y, &length );
    PLA_Obj_local_width   ( Y, &width );
    PLA_Obj_local_buffer  ( Y, &buffer );

    y = (double *)buffer;
    for (i=0; i < length*width; i++) 
        y[i] = ABS( y[i] );

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_abs\n");
    PRISM_print(X);
    PRISM_print(Y);
#endif
}

/* Absolute column sums of a PLAPACK object. */
void PRISM_asum(PLA_Obj X, PLA_Obj Y) 
{
    int i, j, length, width, ldim;
    void *xbuf, *wbuf;
    double *x, *w;
    PLA_Obj W = NULL;

    /* Work space to compute local absolute column sums */
    PLA_Pmvector_create_conf_to( X, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, 1, &W );
    PLA_Obj_set_to_zero(W);

    /* Get local info */
    PLA_Obj_local_length  ( X, &length );
    PLA_Obj_local_width   ( X, &width );
    PLA_Obj_local_ldim    ( X, &ldim );
    PLA_Obj_local_buffer  ( X, &xbuf );
    PLA_Obj_local_buffer  ( W, &wbuf );

    /* Compute local asum values */
    x = (double *)xbuf; 
    w = (double *)wbuf;

    for (j=0; j<width; j++) 
        for (i=0; i<length; i++) 
	        w[j] += ABS( x[i+j*ldim] );

    /* Gather results from other nodes */
    PLA_Reduce( W, MPI_SUM, Y );
    PLA_Obj_free(&W);

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_asum\n");
    PRISM_print(X);
    PRISM_print(Y);
#endif
}

/* Diagonal elements of matrix. */
void PRISM_get_diag(PLA_Obj X, PLA_Obj D) 
{
    int i, width;
	PLA_Obj temp=NULL, t=NULL, x=NULL;

    /* temporary object to hold local diag elements */
    PLA_Mscalar_create_conf_to( D, PLA_ALL_ROWS, PLA_ALL_COLS, &temp );
    PLA_Obj_set_to_zero( temp );

    /* Shift along the diagonal of X */
    PLA_Obj_global_width( X, &width );
    PLA_Obj_view(X, 1, 1, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &x);
    PLA_Obj_view(temp, 1, 1, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &t);
    PLA_Local_copy(x,t); 

    for (i=1; i<width; i++) {
       PLA_Obj_view_shift(x,1,1,1,1); 
       PLA_Obj_view_shift(t,1,0,0,1); 
       PLA_Local_copy(x,t); 
    }

	PLA_Reduce( temp, MPI_SUM, D );

    PLA_Obj_free(&temp);
    PLA_Obj_free(&x);
    PLA_Obj_free(&t);
    
#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_get_diag\n");
    PRISM_print(X);
    PRISM_print(D);
#endif
}

/* 1-norm of a matrix. Method: norm1(x) = max(asum(x)). */
void PRISM_norm1(PLA_Obj X, PLA_Obj alpha) 
{
    int align, length;
    MPI_Datatype datatype; 
    PLA_Template template;
    PLA_Obj temp = NULL;

    PLA_Obj_global_align   ( X, &align );
    PLA_Obj_global_length  ( X, &length );
    PLA_Obj_datatype       ( X, &datatype );
    PLA_Obj_template       ( X, &template );

    PLA_Mvector_create( datatype, length, 1, template, align, &temp );

    PRISM_asum( X, temp );
    PRISM_max( temp, alpha );

    PLA_Obj_free(&temp);

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_norm1\n");
    PRISM_print(X);
    PRISM_print(alpha);
#endif
}

/* Trace of a matrix. Methd: trace(x) = sum(diag(x)). */
void PRISM_trace(PLA_Obj X, PLA_Obj alpha) 
{
    int align, length;
    MPI_Datatype datatype; 
    PLA_Template template;
    PLA_Obj temp = NULL;

    PLA_Obj_global_align   ( X, &align );
    PLA_Obj_global_length  ( X, &length );
    PLA_Obj_datatype       ( X, &datatype );
    PLA_Obj_template       ( X, &template );

    PLA_Mvector_create( datatype, length, 1, template, align, &temp );

    PRISM_get_diag( X, temp );
    PRISM_sum( temp, alpha );

    PLA_Obj_free(&temp);

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_trace\n");
    PRISM_print(X);
    PRISM_print(alpha);
#endif
}

/* Transpose a PLAPACK object */
void PRISM_trans(PLA_Obj X, PLA_Obj T) 
{
    int x_init, t_init;

    PLA_Obj_get_orientation( X, &x_init );
    PLA_Obj_get_orientation( T, &t_init );

    PLA_Obj_set_orientation( X, PLA_PROJ_ONTO_COL );
    PLA_Obj_set_orientation( T, PLA_PROJ_ONTO_ROW );

    PLA_Copy( X, T );

    PLA_Obj_set_orientation( X, x_init );
    PLA_Obj_set_orientation( T, t_init );

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_trans\n");
    PRISM_print(X);
    PRISM_print(T);
#endif
}

/* Set diagonal elements of matrix with a given mvector. */
void PRISM_set_diag(PLA_Obj X, PLA_Obj D) 
{
    int i, width;
    PLA_Obj x = NULL, d = NULL;

    PLA_Obj_set_to_zero( X );

    /* Shift along the diagonal of X */
    PLA_Obj_global_width( X, &width );
    PLA_Obj_view(X, 1, 1, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &x);
    PLA_Obj_view(D, 1, 1, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &d);
    PLA_Copy(d,x); 

    for (i=1; i<width; i++) {
        PLA_Obj_view_shift(x,1,1,1,1); 
        PLA_Obj_view_shift(d,1,0,0,1); 
        PLA_Copy(d,x);
    }

    PLA_Obj_free(&x);
    PLA_Obj_free(&d);

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_set_diag\n");
    PRISM_print(X);
    PRISM_print(D);
#endif
}

/* Create an identity matrix. */
void PRISM_set_to_ident(PLA_Obj X) 
{
    int i, width;
    PLA_Obj x = NULL;

    PLA_Obj_set_to_zero( X );

    /* Shift along the diagonal of X */
    PLA_Obj_global_width( X, &width );
    PLA_Obj_view(X, 1, 1, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &x);
    PLA_Obj_set_to_one(x);

    for (i=1; i<width; i++) {
        PLA_Obj_view_shift(x,1,1,1,1);
        PLA_Obj_set_to_one(x);
    }

    PLA_Obj_free(&x);

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_set_to_ident\n");
    PRISM_print(X);
#endif
}

/* Add a scalar value to a matrix diagonal. */
void PRISM_add_to_diag(PLA_Obj alpha, PLA_Obj X) 
{
    int i, width;
    PLA_Obj x = NULL;

    /* Shift along the diagonal of X */
    PLA_Obj_global_width( X, &width );
    PLA_Obj_view(X, 1, 1, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &x);
    PLA_Local_axpy(PRISM_one, alpha, x);

    for (i=1; i<width; i++) {
        PLA_Obj_view_shift(x,1,1,1,1);
        PLA_Local_axpy(PRISM_one, alpha, x);
    }

    PLA_Obj_free(&x);

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_add_to_diag\n");
    PRISM_print(X);
#endif
}

/* Symmetrize a matrix. Method: X = (X'+X)/2 */
void PRISM_sym(PLA_Obj X) 
{
    double scale = .5;
    MPI_Datatype datatype;
    PLA_Template template;
    PLA_Obj T = NULL, alpha = NULL;

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_sym I\n");
    PRISM_print(X);
#endif

    PLA_Obj_datatype(X, &datatype);
    PLA_Obj_template(X, &template);

    /* temporary objects */
    PLA_Matrix_create_conf_to( X, &T );
    PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, 
      template, &alpha);
    PLA_Obj_set( alpha, datatype, &scale );

    PRISM_trans( X, T );
    PLA_Axpy( PRISM_one, T, X );
    PLA_Scal( alpha, X );

    PLA_Obj_free(&alpha);
    PLA_Obj_free(&T);

#ifdef MISC_DEBUG_
    if (me==ROOT) printf(">> PRISM_sym II\n");
    PRISM_print(X);
#endif
}


/* MISC UTILITIES */

/* Value of a 1x1 MULTISCALAR */
double PRISM_mscalar( PLA_Obj alpha ) {
    void *buffer;
    PLA_Obj_local_buffer( alpha, &buffer );
    return *(double *)buffer;
}

/* Print a global PLAPACK object */
void PRISM_print(PLA_Obj X)
{
    int i, j, k, step = 6;
    int length, width, objtype;
    void *buffer;
    MPI_Datatype datatype;
    PLA_Template template;
    PLA_Obj x=NULL, alpha=NULL, out=NULL;

    /* Extract info from matrix X */
    PLA_Obj_global_length  (X, &length);
    PLA_Obj_global_width   (X, &width);
    PLA_Obj_datatype       (X, &datatype);
    PLA_Obj_template       (X, &template);
    PLA_Obj_objtype        (X, &objtype);

    if (objtype == PLA_MSCALAR) {
        if (me==0) prism_pla_prt_m(X);
        return;
    } 

    PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, 
      template, &alpha);
    PLA_Mscalar_create(datatype, ROOT, ROOT, 1, 1, template, &out);
    PLA_Obj_local_buffer(out, &buffer);

    for (j=0; j<width; j+=step) {
       for (i=0; i<length; i++) 
          for (k=0; k<step && j+k<width; k++) {
             PLA_Obj_view(X, 1, 1, i, j+k, &x); 
             PLA_Obj_set_to_zero(alpha);
             PLA_Local_copy(x,alpha);
             PLA_Reduce(alpha, MPI_SUM, out);
             if (me==ROOT) {
                if (k%step == 0) printf("\n");
                    printf("%12.5e ", *(double *)buffer); 
             }
          }
          if (me==ROOT) printf("\n");
    }
    if (me==ROOT) printf("\n");

    PLA_Obj_free(&x);
    PLA_Obj_free(&out);
    PLA_Obj_free(&alpha);
}

/* Print a global PLAPACK object in long format */
void PRISM_long_print(PLA_Obj X)
{
    int i, j, k, step = 3;
    int length, width, objtype;
    void *buffer;
    MPI_Datatype datatype;
    PLA_Template template;
    PLA_Obj x=NULL, alpha=NULL, out=NULL;

    /* Extract info from matrix X */
    PLA_Obj_global_length  (X, &length);
    PLA_Obj_global_width   (X, &width);
    PLA_Obj_datatype       (X, &datatype);
    PLA_Obj_template       (X, &template);
    PLA_Obj_objtype        (X, &objtype);

    /* deal with (possibly) duplicated objects separately */
    if (objtype == PLA_MSCALAR || objtype == PLA_PVECTOR ||
		objtype == PLA_PMVECTOR) {
        if (me==ROOT) prism_pla_prt_m(X);
        return;
    } 

    PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, 
	  template, &alpha);
    PLA_Mscalar_create(datatype, ROOT, ROOT, 1, 1, template, &out);
    PLA_Obj_local_buffer(out, &buffer);

    for (j=0; j<width; j+=step) {
       for (i=0; i<length; i++) 
          for (k=0; k<step && j+k<width; k++) {
             PLA_Obj_view(X, 1, 1, i, j+k, &x); 
             PLA_Obj_set_to_zero(alpha);
             PLA_Local_copy(x,alpha);
             PLA_Reduce(alpha, MPI_SUM, out);
             if (me==ROOT) {
                if (k%step == 0) printf("\n");
                    printf("%25.15e ", *(double *)buffer); 
             }
          }
          if (me==ROOT) printf("\n");
    }
    if (me==ROOT) printf("\n");

    PLA_Obj_free(&x);
    PLA_Obj_free(&out);
    PLA_Obj_free(&alpha);
}
